# Filename: test_tools.py
# pylint: disable=locally-disabled,C0111,R0904,C0103

from io import StringIO

import numpy as np

from km3pipe.testing import TestCase, patch
from km3pipe.tools import (
    unpack_nfirst,
    split,
    namedtuple_with_defaults,
    remain_file_pointer,
    decamelise,
    camelise,
    issorted,
    lstrip,
    chunks,
    is_coherent,
    istype,
    get_jpp_version,
)

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
__status__ = "Development"


class TestTools(TestCase):
    def setUp(self):
        self.vecs = np.array(
            [
                [0.0, 1.0, 5.0],
                [1.0, 1.0, 4.0],
                [2.0, 1.0, 3.0],
                [3.0, 1.0, 2.0],
                [4.0, 1.0, 1.0],
            ]
        )
        self.v = (1, 2, 3)
        self.unit_v = np.array([0.26726124, 0.53452248, 0.80178373])
        self.unit_vecs = np.array(
            [
                [0.0, 0.19611614, 0.98058068],
                [0.23570226, 0.23570226, 0.94280904],
                [0.53452248, 0.26726124, 0.80178373],
                [0.80178373, 0.26726124, 0.53452248],
                [0.94280904, 0.23570226, 0.23570226],
            ]
        )

    def test_unpack_nfirst(self):
        a_tuple = (1, 2, 3, 4, 5)
        a, b, c, rest = unpack_nfirst(a_tuple, 3)
        self.assertEqual(1, a)
        self.assertEqual(2, b)
        self.assertEqual(3, c)
        self.assertTupleEqual((4, 5), rest)

    def test_split_splits_strings(self):
        string = "1 2 3 4"
        parts = split(string)
        self.assertListEqual(["1", "2", "3", "4"], parts)

    def test_split_splits_strings_with_separator(self):
        string = "1,2,3,4"
        parts = split(string, sep=",")
        self.assertListEqual(["1", "2", "3", "4"], parts)

    def test_split_callback_converts_correctly(self):
        string = "1 2 3 4"
        parts = split(string, int)
        self.assertListEqual([1, 2, 3, 4], parts)

        string = "1.0 2.1 3.2 4.3"
        parts = split(string, float)
        self.assertListEqual([1.0, 2.1, 3.2, 4.3], parts)

    def test_namedtuple_with_defaults_initialises_with_none(self):
        Node = namedtuple_with_defaults("Node", "val left right")
        node = Node()
        self.assertIsNone(node.val)
        self.assertIsNone(node.left)
        self.assertIsNone(node.right)

    def test_namedtuple_with_defaults_initialises_with_given_values(self):
        Node = namedtuple_with_defaults("Node", "val left right", [1, 2, 3])
        node = Node()
        self.assertEqual(1, node.val)
        self.assertEqual(2, node.left)
        self.assertEqual(3, node.right)


class TestRemainFilePointer(TestCase):
    def test_remains_file_pointer_in_function(self):
        dummy_file = StringIO("abcdefg")

        @remain_file_pointer
        def seek_into_file(file_obj):
            file_obj.seek(1, 0)

        dummy_file.seek(2, 0)
        self.assertEqual(2, dummy_file.tell())
        seek_into_file(dummy_file)
        self.assertEqual(2, dummy_file.tell())

    def test_remains_file_pointer_and_return_value_in_function(self):
        dummy_file = StringIO("abcdefg")

        @remain_file_pointer
        def seek_into_file(file_obj):
            file_obj.seek(1, 0)
            return 1

        dummy_file.seek(2, 0)
        self.assertEqual(2, dummy_file.tell())
        return_value = seek_into_file(dummy_file)
        self.assertEqual(2, dummy_file.tell())
        self.assertEqual(1, return_value)

    def test_remains_file_pointer_in_class_method(self):
        class FileSeekerClass(object):
            def __init__(self):
                self.dummy_file = StringIO("abcdefg")

            @remain_file_pointer
            def seek_into_file(self, file_obj):
                file_obj.seek(1, 0)

        fileseeker = FileSeekerClass()
        fileseeker.dummy_file.seek(2, 0)
        self.assertEqual(2, fileseeker.dummy_file.tell())
        fileseeker.seek_into_file(fileseeker.dummy_file)
        self.assertEqual(2, fileseeker.dummy_file.tell())

    def test_remains_file_pointer_and_return_value_in_class_method(self):
        class FileSeekerClass(object):
            def __init__(self):
                self.dummy_file = StringIO("abcdefg")

            @remain_file_pointer
            def seek_into_file(self, file_obj):
                file_obj.seek(1, 0)
                return 1

        fileseeker = FileSeekerClass()
        fileseeker.dummy_file.seek(2, 0)
        self.assertEqual(2, fileseeker.dummy_file.tell())
        return_value = fileseeker.seek_into_file(fileseeker.dummy_file)
        self.assertEqual(2, fileseeker.dummy_file.tell())
        self.assertEqual(1, return_value)


class TestCamelCaseConverter(TestCase):
    def test_decamelise(self):
        text = "TestCase"
        self.assertEqual("test_case", decamelise(text))
        text = "TestCaseXYZ"
        self.assertEqual("test_case_xyz", decamelise(text))
        text = "1TestCase"
        self.assertEqual("1_test_case", decamelise(text))
        text = "test_case"
        self.assertEqual("test_case", decamelise(text))

    def test_camelise(self):
        text = "camel_case"
        self.assertEqual("CamelCase", camelise(text))
        text = "camel_case"
        self.assertEqual("camelCase", camelise(text, capital_first=False))


class TestMisc(TestCase):
    def test_issorted(self):
        assert issorted([1, 2, 3])
        assert not issorted([2, 3, 1])

    def test_is_coherent(self):
        assert is_coherent([1, 2, 3, 4, 5])
        assert not is_coherent([1, 3, 4, 5])


class TestLstrip(TestCase):
    def test_lstrip(self):
        text = """
           a
            b
             c
            d
          e
        """
        self.assertEqual("a\nb\nc\nd\ne\n", lstrip(text))

    def test_lstrip_on_single_line(self):
        text = "  a  "
        self.assertEqual(text.lstrip(), lstrip(text))


class TestChunks(TestCase):
    def test_chunks(self):
        l = [1, 2, 3, 4, 5, 6, 7, 8, 9]
        self.assertEqual([[1, 2, 3], [4, 5, 6], [7, 8, 9]], list(chunks(l, 3)))
        self.assertEqual([[1, 2, 3, 4], [5, 6, 7, 8], [9]], list(chunks(l, 4)))


class TestIstype(TestCase):
    def test_a_type(self):
        a = [1]
        assert istype(a, "list")

    def test_another_type(self):
        b = "string"
        if isinstance(b, str):
            assert istype(b, "str")
        else:
            assert istype(b, "unicode")


class TestJppRevision(TestCase):
    @patch("subprocess.check_output")
    def test_version(self, co_mock):
        co_mock.return_value = b"version:    8519\nname space: KM3NET\n"
        assert "8519" == get_jpp_version()

    def test_version(self):
        assert get_jpp_version(via_command="a_non_existing_command") is None



# Filename: hdf5.py
# pylint: disable=C0103,R0903,C901
# vim:set ts=4 sts=4 sw=4 et:
"""
Read and write KM3NeT-formatted HDF5 files.

"""

from collections import OrderedDict, defaultdict, namedtuple
import os.path
import warnings

import numpy as np
import tables as tb
import km3io
from thepipe import Provenance

try:
    from numba import jit
except ImportError:
    jit = lambda f: f

import km3pipe as kp
from km3pipe.core import Pump, Module, Blob
from km3pipe.dataclasses import Table, NDArray
from km3pipe.logger import get_logger
from km3pipe.tools import decamelise, camelise, split, istype, get_jpp_version

log = get_logger(__name__)  # pylint: disable=C0103

__author__ = "Tamas Gal and Moritz Lotze and Michael Moser"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal and Moritz Lotze"
__email__ = "tgal@km3net.de"
__status__ = "Development"

FORMAT_VERSION = np.string_("5.1")
MINIMUM_FORMAT_VERSION = np.string_("4.1")


class H5VersionError(Exception):
    pass


def check_version(h5file):
    try:
        version = np.string_(h5file.root._v_attrs.format_version)
    except AttributeError:
        log.error(
            "Could not determine HDF5 format version: '%s'."
            "You may encounter unexpected errors! Good luck..." % h5file.filename
        )
        return
    if split(version, int, np.string_(".")) < split(
        MINIMUM_FORMAT_VERSION, int, np.string_(".")
    ):
        raise H5VersionError(
            "HDF5 format version {0} or newer required!\n"
            "'{1}' has HDF5 format version {2}.".format(
                MINIMUM_FORMAT_VERSION.decode("utf-8"),
                h5file.filename,
                version.decode("utf-8"),
            )
        )


class HDF5Header(object):
    """Wrapper class for the `/raw_header` table in KM3HDF5"""

    def __init__(self, data):
        self._data = data
        self._set_attributes()

    def _set_attributes(self):
        """Traverse the internal dictionary and set the getters"""
        for parameter, data in self._data.items():
            if isinstance(data, dict) or isinstance(data, OrderedDict):
                field_names, field_values = zip(*data.items())
                sorted_indices = np.argsort(field_names)
                attr = namedtuple(parameter, [field_names[i] for i in sorted_indices])
                setattr(
                    self, parameter, attr(*[field_values[i] for i in sorted_indices])
                )
            else:
                setattr(self, parameter, data)

    @classmethod
    def from_table(cls, table):
        data = OrderedDict()
        for i in range(len(table)):
            parameter = table["parameter"][i]
            field_names = table["field_names"][i].split(" ")
            field_values = table["field_values"][i].split(" ")
            if field_values == [""]:
                log.info("No value for parameter '{}'! Skipping...".format(parameter))
                continue
            dtypes = table["dtype"][i]
            dtyped_values = []
            for dtype, value in zip(dtypes.split(" "), field_values):
                if dtype.startswith("a"):
                    dtyped_values.append(value)
                else:
                    value = np.fromstring(value, dtype=dtype, sep=" ")[0]
                    dtyped_values.append(value)
            data[parameter] = OrderedDict(zip(field_names, dtyped_values))
        return cls(data)

    @classmethod
    def from_km3io(cls, header):
        if not isinstance(header, km3io.offline.Header):
            raise TypeError(
                "The given header object is not an instance of km3io.offline.Header"
            )
        h5header = cls({})
        for attribute, value in header._data.items():
            setattr(h5header, attribute, value)
        return h5header

    @classmethod
    def from_aanet(cls, table):
        data = OrderedDict()
        for i in range(len(table)):
            parameter = table["parameter"][i].astype(str)
            field_names = [n.decode() for n in table["field_names"][i].split()]
            field_values = [n.decode() for n in table["field_values"][i].split()]
            if field_values in [[b""], []]:
                log.info("No value for parameter '{}'! Skipping...".format(parameter))
                continue
            dtypes = table["dtype"][i]
            dtyped_values = []
            for dtype, value in zip(dtypes.split(), field_values):
                if dtype.startswith(b"a"):
                    dtyped_values.append(value)
                else:
                    value = np.fromstring(value, dtype=dtype, sep=" ")[0]
                    dtyped_values.append(value)
            data[parameter] = OrderedDict(zip(field_names, dtyped_values))
        return cls(data)

    @classmethod
    def from_hdf5(cls, filename):
        with tb.open_file(filename, "r") as f:
            table = f.get_node("/raw_header")
            return cls.from_pytable(table)

    @classmethod
    def from_pytable(cls, table):
        data = OrderedDict()
        for row in table:
            parameter = row["parameter"].decode()
            field_names = row["field_names"].decode().split(" ")
            field_values = row["field_values"].decode().split(" ")
            if field_values == [""]:
                log.info("No value for parameter '{}'! Skipping...".format(parameter))
                continue
            dtypes = row["dtype"].decode()
            dtyped_values = []
            for dtype, value in zip(dtypes.split(" "), field_values):
                if dtype.startswith("a"):
                    dtyped_values.append(value)
                else:
                    value = np.fromstring(value, dtype=dtype, sep=" ")[0]
                    dtyped_values.append(value)
            data[parameter] = OrderedDict(zip(field_names, dtyped_values))
        return cls(data)


class HDF5IndexTable(object):
    def __init__(self, h5loc):
        self.h5loc = h5loc
        self._data = defaultdict(list)
        self._index = 0

    def append(self, n_items):
        self._data["indices"].append(self._index)
        self._data["n_items"].append(n_items)
        self._index += n_items

    @property
    def data(self):
        return self._data


class HDF5Sink(Module):
    """Write KM3NeT-formatted HDF5 files, event-by-event.

    The data can be a ``kp.Table``, a numpy structured array,
    a pandas DataFrame, or a simple scalar.

    The name of the corresponding H5 table is the decamelised
    blob-key, so values which are stored in the blob under `FooBar`
    will be written to `/foo_bar` in the HDF5 file.

    Parameters
    ----------
    filename: str, optional [default: 'dump.h5']
        Where to store the events.
    h5file: pytables.File instance, optional [default: None]
        Opened file to write to. This is mutually exclusive with filename.
    keys: list of strings, optional
        List of Blob-keys to write, everything else is ignored.
    complib : str [default: zlib]
        Compression library that should be used.
        'zlib', 'lzf', 'blosc' and all other PyTables filters
        are available.
    complevel : int [default: 5]
        Compression level.
    chunksize : int [optional]
        Chunksize that should be used for saving along the first axis
        of the input array.
    flush_frequency: int, optional [default: 500]
        The number of iterations to cache tables and arrays before
        dumping to disk.
    pytab_file_args: dict [optional]
        pass more arguments to the pytables File init
    n_rows_expected = int, optional [default: 10000]
    append: bool, optional [default: False]

    Services
    --------
    write_table: Table
        The table to write, with "h5loc" set

    """

    def configure(self):
        self.filename = self.get("filename", default="dump.h5")
        self.ext_h5file = self.get("h5file")
        self.keys = self.get("keys", default=[])
        self.complib = self.get("complib", default="zlib")
        self.complevel = self.get("complevel", default=5)
        self.chunksize = self.get("chunksize")
        self.flush_frequency = self.get("flush_frequency", default=500)
        self.pytab_file_args = self.get("pytab_file_args", default=dict())
        self.file_mode = "a" if self.get("append") else "w"
        self.keep_open = self.get("keep_open")
        self.indices = {}  # to store HDF5IndexTables for each h5loc
        self._singletons_written = {}
        # magic 10000: this is the default of the "expectedrows" arg
        # from the tables.File.create_table() function
        # at least according to the docs
        # might be able to set to `None`, I don't know...
        self.n_rows_expected = self.get("n_rows_expected", default=10000)
        self.index = 0

        self.expose(self.write_table, "write_table")

        if self.ext_h5file is not None:
            self.h5file = self.ext_h5file
        else:
            self.h5file = tb.open_file(
                self.filename,
                mode=self.file_mode,
                title="KM3NeT",
                **self.pytab_file_args
            )
            Provenance().record_output(self.filename, "HDF5Sink output")
        self.filters = tb.Filters(
            complevel=self.complevel,
            shuffle=True,
            fletcher32=True,
            complib=self.complib,
        )
        self._tables = OrderedDict()
        self._ndarrays = OrderedDict()
        self._ndarrays_cache = defaultdict(list)

    def _to_array(self, data, name=None):
        if data is None:
            return
        if np.isscalar(data):
            self.log.debug("toarray: is a scalar")
            return Table(
                {name: np.asarray(data).reshape((1,))},
                h5loc="/misc/{}".format(decamelise(name)),
                name=name,
            )
        if hasattr(data, "len") and len(data) <= 0:  # a bit smelly ;)
            self.log.debug("toarray: data has no length")
            return
        # istype instead isinstance, to avoid heavy pandas import (hmmm...)
        if istype(data, "DataFrame"):  # noqa
            self.log.debug("toarray: pandas dataframe")
            data = Table.from_dataframe(data)
        return data

    def _cache_ndarray(self, arr):
        self._ndarrays_cache[arr.h5loc].append(arr)

    def _write_ndarrays_cache_to_disk(self):
        """Writes all the cached NDArrays to disk and empties the cache"""
        for h5loc, arrs in self._ndarrays_cache.items():
            title = arrs[0].title
            chunkshape = (
                (self.chunksize,) + arrs[0].shape[1:]
                if self.chunksize is not None
                else None
            )

            arr = NDArray(np.concatenate(arrs), h5loc=h5loc, title=title)

            if h5loc not in self._ndarrays:
                loc, tabname = os.path.split(h5loc)
                ndarr = self.h5file.create_earray(
                    loc,
                    tabname,
                    tb.Atom.from_dtype(arr.dtype),
                    (0,) + arr.shape[1:],
                    chunkshape=chunkshape,
                    title=title,
                    filters=self.filters,
                    createparents=True,
                )
                self._ndarrays[h5loc] = ndarr
            else:
                ndarr = self._ndarrays[h5loc]

            idx_table_h5loc = h5loc + "_indices"
            if idx_table_h5loc not in self.indices:
                self.indices[idx_table_h5loc] = HDF5IndexTable(idx_table_h5loc)
            idx_tab = self.indices[idx_table_h5loc]

            for arr_length in (len(a) for a in arrs):
                idx_tab.append(arr_length)

            ndarr.append(arr)

        self._ndarrays_cache = defaultdict(list)

    def write_table(self, table):
        """Write a single table to the HDF5 file, exposed as a service"""
        self._write_table(table.h5loc, table, table.name)

    def _write_table(self, h5loc, arr, title):
        level = len(h5loc.split("/"))

        if h5loc not in self._tables:
            dtype = arr.dtype
            if any("U" in str(dtype.fields[f][0]) for f in dtype.fields):
                self.log.error(
                    "Cannot write data to '{}'. Unicode strings are not supported!".format(
                        h5loc
                    )
                )
                return
            loc, tabname = os.path.split(h5loc)
            self.log.debug(
                "h5loc '{}', Loc '{}', tabname '{}'".format(h5loc, loc, tabname)
            )
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", tb.NaturalNameWarning)
                tab = self.h5file.create_table(
                    loc,
                    tabname,
                    chunkshape=self.chunksize,
                    description=dtype,
                    title=title,
                    filters=self.filters,
                    createparents=True,
                    expectedrows=self.n_rows_expected,
                )
            tab._v_attrs.datatype = title
            if level < 5:
                self._tables[h5loc] = tab
        else:
            tab = self._tables[h5loc]

        h5_colnames = set(tab.colnames)
        tab_colnames = set(arr.dtype.names)
        if h5_colnames != tab_colnames:
            missing_cols = h5_colnames - tab_colnames
            if missing_cols:
                self.log.info("Missing columns in table, trying to append NaNs.")
                arr = arr.append_columns(
                    missing_cols, np.full((len(missing_cols), len(arr)), np.nan)
                )
                if arr.dtype != tab.dtype:
                    self.log.error(
                        "Differing dtypes after appending "
                        "missing columns to the table! Skipping..."
                    )
                    return

        if arr.dtype != tab.dtype:
            try:
                arr = Table(arr, dtype=tab.dtype)
            except ValueError:
                self.log.critical(
                    "Cannot write a table to '%s' since its dtype is "
                    "different compared to the previous table with the same "
                    "HDF5 location, which was used to fix the dtype of the "
                    "HDF5 compund type." % h5loc
                )
                raise

        tab.append(arr)

        if level < 4:
            tab.flush()

    def _write_separate_columns(self, where, obj, title):
        f = self.h5file
        loc, group_name = os.path.split(where)
        if where not in f:
            group = f.create_group(loc, group_name, title, createparents=True)
            group._v_attrs.datatype = title
        else:
            group = f.get_node(where)

        for col, (dt, _) in obj.dtype.fields.items():
            data = obj.__array__()[col]

            if col not in group:
                a = tb.Atom.from_dtype(dt)
                arr = f.create_earray(
                    group, col, a, (0,), col.capitalize(), filters=self.filters
                )
            else:
                arr = getattr(group, col)
            arr.append(data)

        # create index table
        if where not in self.indices:
            self.indices[where] = HDF5IndexTable(where + "/_indices")
        idx_tab = self.indices[where]
        idx_tab.append(len(data))

    def _process_entry(self, key, entry):
        self.log.debug("Inspecting {}".format(key))
        if (
            hasattr(entry, "h5singleton")
            and entry.h5singleton
            and entry.h5loc in self._singletons_written
        ):
            self.log.debug(
                "Skipping '%s' since it's a singleton and already written."
                % entry.h5loc
            )
            return
        if not hasattr(entry, "h5loc"):
            self.log.debug("Ignoring '%s': no h5loc attribute" % key)
            return
        if isinstance(entry, NDArray):
            self._cache_ndarray(entry)
            return entry
        try:
            title = entry.name
        except AttributeError:
            title = key

        if isinstance(entry, Table) and not entry.h5singleton:
            if "group_id" not in entry:
                entry = entry.append_columns("group_id", self.index)

        self.log.debug("h5l: '{}', title '{}'".format(entry.h5loc, title))

        if hasattr(entry, "split_h5") and entry.split_h5:
            self.log.debug("Writing into separate columns...")
            self._write_separate_columns(entry.h5loc, entry, title=title)
        else:
            self.log.debug("Writing into single Table...")
            self._write_table(entry.h5loc, entry, title=title)

        if hasattr(entry, "h5singleton") and entry.h5singleton:
            self._singletons_written[entry.h5loc] = True

        return entry

    def process(self, blob):
        written_blob = Blob()
        for key, entry in sorted(blob.items()):
            if self.keys and key not in self.keys:
                self.log.info("Skipping blob, since it's not in the keys list")
                continue
            data = self._process_entry(key, entry)
            if data is not None:
                written_blob[key] = data

        if "GroupInfo" not in blob:
            gi = Table(
                {"group_id": self.index, "blob_length": len(written_blob)},
                h5loc="/group_info",
                name="Group Info",
            )
            self._process_entry("GroupInfo", gi)

        if not self.index % self.flush_frequency:
            self.flush()

        self.index += 1
        return blob

    def flush(self):
        """Flush tables and arrays to disk"""
        self.log.info("Flushing tables and arrays to disk...")
        for tab in self._tables.values():
            tab.flush()
        self._write_ndarrays_cache_to_disk()

    def finish(self):
        self.flush()
        self.h5file.root._v_attrs.km3pipe = np.string_(kp.__version__)
        self.h5file.root._v_attrs.pytables = np.string_(tb.__version__)
        self.h5file.root._v_attrs.jpp = np.string_(get_jpp_version())
        self.h5file.root._v_attrs.format_version = np.string_(FORMAT_VERSION)
        self.log.info("Adding index tables.")
        for where, idx_tab in self.indices.items():
            self.log.debug("Creating index table for '%s'" % where)
            h5loc = idx_tab.h5loc
            self.log.info("  -> {0}".format(h5loc))
            indices = Table(
                {"index": idx_tab.data["indices"], "n_items": idx_tab.data["n_items"]},
                h5loc=h5loc,
            )
            self._write_table(h5loc, indices, title="Indices")
        self.log.info(
            "Creating pytables index tables. " "This may take a few minutes..."
        )
        for tab in self._tables.values():
            if "frame_id" in tab.colnames:
                tab.cols.frame_id.create_index()
            if "slice_id" in tab.colnames:
                tab.cols.slice_id.create_index()
            if "dom_id" in tab.colnames:
                tab.cols.dom_id.create_index()
            if "event_id" in tab.colnames:
                try:
                    tab.cols.event_id.create_index()
                except NotImplementedError:
                    log.warning(
                        "Table '{}' has an uint64 column, "
                        "not indexing...".format(tab._v_name)
                    )
            if "group_id" in tab.colnames:
                try:
                    tab.cols.group_id.create_index()
                except NotImplementedError:
                    log.warning(
                        "Table '{}' has an uint64 column, "
                        "not indexing...".format(tab._v_name)
                    )
            tab.flush()

        if "HDF5MetaData" in self.services:
            self.log.info("Writing HDF5 meta data.")
            metadata = self.services["HDF5MetaData"]
            for name, value in metadata.items():
                self.h5file.set_node_attr("/", name, value)

        if not self.keep_open:
            self.h5file.close()
        self.cprint("HDF5 file written to: {}".format(self.filename))


class HDF5Pump(Pump):
    """Read KM3NeT-formatted HDF5 files, event-by-event.

    Parameters
    ----------
    filename: str
        From where to read events. Either this OR ``filenames`` needs to be
        defined.
    skip_version_check: bool [default: False]
        Don't check the H5 version. Might lead to unintended consequences.
    shuffle: bool, optional [default: False]
        Shuffle the group_ids, so that the blobs are mixed up.
    shuffle_function: function, optional [default: np.random.shuffle
        The function to be used to shuffle the group IDs.
    reset_index: bool, optional [default: True]
        When shuffle is set to true, reset the group ID - start to count
        the group_id by 0.
    """

    def configure(self):
        self.filename = self.get("filename")
        self.skip_version_check = self.get("skip_version_check", default=False)
        self.verbose = bool(self.get("verbose"))
        self.shuffle = self.get("shuffle", default=False)
        self.shuffle_function = self.get("shuffle_function", default=np.random.shuffle)
        self.reset_index = self.get("reset_index", default=False)

        self.h5file = None
        self.cut_mask = None
        self.indices = {}
        self._tab_indices = {}
        self._singletons = {}
        self.header = None
        self.group_ids = None
        self._n_groups = None
        self.index = 0

        self.h5file = tb.open_file(self.filename, "r")

        Provenance().record_input(self.filename, "HDF5Pump input")

        if not self.skip_version_check:
            check_version(self.h5file)

        self._read_group_info()

    def _read_group_info(self):
        h5file = self.h5file

        if "/group_info" not in h5file:
            self.log.critical("Missing /group_info '%s', aborting..." % h5file.filename)
            raise SystemExit

        self.log.info("Reading group information from '/group_info'.")
        group_info = h5file.get_node("/", "group_info")
        self.group_ids = group_info.cols.group_id[:]
        self._n_groups = len(self.group_ids)

        if "/raw_header" in h5file:
            self.log.info("Reading /raw_header")
            try:
                self.header = HDF5Header.from_pytable(h5file.get_node("/raw_header"))
            except TypeError:
                self.log.error("Could not parse the raw header, skipping!")

        if self.shuffle:
            self.log.info("Shuffling group IDs")
            self.shuffle_function(self.group_ids)

    def process(self, blob):
        self.log.info("Reading blob at index %s" % self.index)
        if self.index >= self._n_groups:
            self.log.info("All groups are read.")
            raise StopIteration
        blob = self.get_blob(self.index)
        self.index += 1
        return blob

    def get_blob(self, index):
        blob = Blob()
        group_id = self.group_ids[index]

        # skip groups with separate columns
        # and deal with them later
        # this should be solved using hdf5 attributes in near future
        split_table_locs = []
        ndarray_locs = []
        for tab in self.h5file.walk_nodes(classname="Table"):
            h5loc = tab._v_pathname
            loc, tabname = os.path.split(h5loc)
            if tabname in self.indices:
                self.log.info("index table '%s' already read, skip..." % h5loc)
                continue
            if loc in split_table_locs:
                self.log.info("get_blob: '%s' is noted, skip..." % h5loc)
                continue
            if tabname == "_indices":
                self.log.debug("get_blob: found index table '%s'" % h5loc)
                split_table_locs.append(loc)
                self.indices[loc] = self.h5file.get_node(h5loc)
                continue
            if tabname.endswith("_indices"):
                self.log.debug("get_blob: found index table '%s' for NDArray" % h5loc)
                ndarr_loc = h5loc.replace("_indices", "")
                ndarray_locs.append(ndarr_loc)
                if ndarr_loc in self.indices:
                    self.log.info(
                        "index table for NDArray '%s' already read, skip..." % ndarr_loc
                    )
                    continue
                _index_table = self.h5file.get_node(h5loc)
                self.indices[ndarr_loc] = {
                    "index": _index_table.col("index")[:],
                    "n_items": _index_table.col("n_items")[:],
                }
                continue
            tabname = camelise(tabname)

            if "group_id" in tab.dtype.names:
                try:
                    if h5loc not in self._tab_indices:
                        self._read_tab_indices(h5loc)
                    tab_idx_start = self._tab_indices[h5loc][0][group_id]
                    tab_n_items = self._tab_indices[h5loc][1][group_id]
                    if tab_n_items == 0:
                        continue
                    arr = tab[tab_idx_start : tab_idx_start + tab_n_items]
                except IndexError:
                    self.log.debug("No data for h5loc '%s'" % h5loc)
                    continue
                except NotImplementedError:
                    # 64-bit unsigned integer columns like ``group_id``
                    # are not yet supported in conditions
                    self.log.debug(
                        "get_blob: found uint64 column at '{}'...".format(h5loc)
                    )
                    arr = tab.read()
                    arr = arr[arr["group_id"] == group_id]
                except ValueError:
                    # "there are no columns taking part
                    # in condition ``group_id == 0``"
                    self.log.info(
                        "get_blob: no `%s` column found in '%s'! "
                        "skipping... " % ("group_id", h5loc)
                    )
                    continue
            else:
                if h5loc not in self._singletons:
                    log.info("Caching H5 singleton: {} ({})".format(tabname, h5loc))
                    self._singletons[h5loc] = Table(
                        tab.read(),
                        h5loc=h5loc,
                        split_h5=False,
                        name=tabname,
                        h5singleton=True,
                    )
                blob[tabname] = self._singletons[h5loc]
                continue

            self.log.debug("h5loc: '{}'".format(h5loc))
            tab = Table(arr, h5loc=h5loc, split_h5=False, name=tabname)
            if self.shuffle and self.reset_index:
                tab.group_id[:] = index
            blob[tabname] = tab

        # skipped locs are now column wise datasets (usually hits)
        # currently hardcoded, in future using hdf5 attributes
        # to get the right constructor
        for loc in split_table_locs:
            # if some events are missing (group_id not continuous),
            # this does not work as intended
            # idx, n_items = self.indices[loc][group_id]
            idx = self.indices[loc].col("index")[group_id]
            n_items = self.indices[loc].col("n_items")[group_id]
            end = idx + n_items
            node = self.h5file.get_node(loc)
            columns = (c for c in node._v_children if c != "_indices")
            data = {}
            for col in columns:
                data[col] = self.h5file.get_node(loc + "/" + col)[idx:end]
            tabname = camelise(loc.split("/")[-1])
            s_tab = Table(data, h5loc=loc, split_h5=True, name=tabname)
            if self.shuffle and self.reset_index:
                s_tab.group_id[:] = index
            blob[tabname] = s_tab

        if self.header is not None:
            blob["Header"] = self.header

        for ndarr_loc in ndarray_locs:
            self.log.info("Reading %s" % ndarr_loc)
            try:
                idx = self.indices[ndarr_loc]["index"][group_id]
                n_items = self.indices[ndarr_loc]["n_items"][group_id]
            except IndexError:
                continue
            end = idx + n_items
            ndarr = self.h5file.get_node(ndarr_loc)
            ndarr_name = camelise(ndarr_loc.split("/")[-1])
            _ndarr = NDArray(
                ndarr[idx:end], h5loc=ndarr_loc, title=ndarr.title, group_id=group_id
            )
            if self.shuffle and self.reset_index:
                _ndarr.group_id = index
            blob[ndarr_name] = _ndarr

        return blob

    def _read_tab_indices(self, h5loc):
        self.log.info("Reading table indices for '{}'".format(h5loc))
        node = self.h5file.get_node(h5loc)
        group_ids = None
        if "group_id" in node.dtype.names:
            group_ids = self.h5file.get_node(h5loc).cols.group_id[:]
        else:
            self.log.error("No data found in '{}'".format(h5loc))
            return

        self._tab_indices[h5loc] = create_index_tuple(group_ids)

    def __len__(self):
        self.log.info("Opening all HDF5 files to check the number of groups")
        n_groups = 0
        for filename in self.filenames:
            with tb.open_file(filename, "r") as h5file:
                group_info = h5file.get_node("/", "group_info")
                self.group_ids = group_info.cols.group_id[:]
                n_groups += len(self.group_ids)
        return n_groups

    def __iter__(self):
        return self

    def __next__(self):
        # TODO: wrap that in self._check_if_next_file_is_needed(self.index)
        if self.index >= self._n_groups:
            self.log.info("All groups are read")
            raise StopIteration
        blob = self.get_blob(self.index)
        self.index += 1
        return blob

    def __getitem__(self, index):
        if isinstance(index, int):
            return self.get_blob(index)
        elif isinstance(index, slice):
            return self._slice_generator(index)
        else:
            raise TypeError("index must be int or slice")

    def _slice_generator(self, index):
        """A simple slice generator for iterations"""
        start, stop, step = index.indices(len(self))
        for i in range(start, stop, step):
            yield self.get_blob(i)

        self.filename = None

    def _close_h5file(self):
        if self.h5file:
            self.h5file.close()

    def finish(self):
        self._close_h5file()


@jit
def create_index_tuple(group_ids):
    """An helper function to create index tuples for fast lookup in HDF5Pump"""
    max_group_id = np.max(group_ids)

    start_idx_arr = np.full(max_group_id + 1, 0)
    n_items_arr = np.full(max_group_id + 1, 0)

    current_group_id = group_ids[0]
    current_idx = 0
    item_count = 0

    for group_id in group_ids:
        if group_id != current_group_id:
            start_idx_arr[current_group_id] = current_idx
            n_items_arr[current_group_id] = item_count
            current_idx += item_count
            item_count = 0
            current_group_id = group_id
        item_count += 1
    else:
        start_idx_arr[current_group_id] = current_idx
        n_items_arr[current_group_id] = item_count

    return (start_idx_arr, n_items_arr)


class HDF5MetaData(Module):
    """Metadata to attach to the HDF5 file.

    Parameters
    ----------
    data: dict

    """

    def configure(self):
        self.data = self.require("data")
        self.expose(self.data, "HDF5MetaData")


def convert_header_dict_to_table(header_dict):
    """Converts a header dictionary (usually from aanet) to a Table"""
    if not header_dict:
        log.warning("Can't convert empty header dict to table, skipping...")
        return
    tab_dict = defaultdict(list)
    log.debug("Param:   field_names    field_values    dtype")
    for parameter, data in header_dict.items():
        fields = []
        values = []
        types = []
        for field_name, field_value in data.items():
            fields.append(field_name)
            values.append(str(field_value))
            try:
                _ = float(field_value)  # noqa
                types.append("f4")
            except ValueError:
                types.append("a{}".format(len(field_value)))
        tab_dict["parameter"].append(parameter)
        tab_dict["field_names"].append(" ".join(fields))
        tab_dict["field_values"].append(" ".join(values))
        tab_dict["dtype"].append(" ".join(types))
        log.debug(
            "{}: {} {} {}".format(
                tab_dict["parameter"][-1],
                tab_dict["field_names"][-1],
                tab_dict["field_values"][-1],
                tab_dict["dtype"][-1],
            )
        )
    return Table(tab_dict, h5loc="/raw_header", name="RawHeader", h5singleton=True)



#!/usr/bin/env python

from time import ctime

from royweb import PacketHandler

from km3pipe import Pipeline, Module
from km3pipe.io import CHPump
from km3pipe.io.daq import DAQPreamble, DAQEvent
from io import StringIO


class CHPrinter(Module):
    def process(self, blob):
        print("New blob:")
        print(blob["CHPrefix"])
        return blob


class ROySender(Module):
    def configure(self):
        self.packet_handler = PacketHandler("131.188.161.241", 9999)

    def process(self, blob):
        roy = self.packet_handler
        data_size = blob["CHPrefix"].length
        roy.send("evt_size", data_size, "byte", "")

        data = blob["CHData"]
        data_io = StringIO(data)

        preamble = DAQPreamble(file_obj=data_io)
        print(ctime())
        print(preamble)

        event = DAQEvent(file_obj=data_io)
        print("Run: {0}".format(event.header.run))
        print("Time Slice: {0}".format(event.header.time_slice))

        try:
            ratio = event.n_triggered_hits / event.n_snapshot_hits * 100
        except ZeroDivisionError:
            ratio = 100 if event.n_triggered_hits > 1 else 0
        roy.send("hit_ratio", ratio, "%", "")
        roy.send("trig_hits", event.n_triggered_hits, "#", "")
        roy.send("hits", event.n_snapshot_hits, "#", "")
        roy.send("timeslice", event.header.time_slice, "#", "")
        roy.send("run", event.header.run, "#", "")

        doms = {dom_id for (dom_id, _, _, _) in event.snapshot_hits}
        trig_doms = {dom_id for (dom_id, _, _, _, _) in event.triggered_hits}
        roy.send("doms", len(doms), "#", "")
        roy.send("trig_doms", len(trig_doms), "#", "")

        hit_times = [time for (_, _, time, _) in event.snapshot_hits]
        try:
            length = abs(max(hit_times) - min(hit_times))
        except ValueError:
            pass
        else:
            roy.send("evt_length", length, "ns", "")

        # if event.n_snapshot_hits < 30:
        #    tots = [tot for (_, _, _, tot) in event.snapshot_hits]
        #    for tot in tots[::3]:
        #        roy.send("tot", tot, 'ns', '')

        return blob


pipe = Pipeline()
pipe.attach(
    CHPump,
    host="localhost",
    port=5553,
    tags="IO_EVT",
    timeout=60 * 60 * 24,
    max_queue=50,
)
pipe.attach(ROySender)
pipe.attach(CHPrinter)

pipe.drain()



# Filename: shell.py
# pylint: disable=C0103
"""
Some shell helpers

"""
import os
import subprocess

from .tools import lstrip
from .logger import get_logger

from subprocess import DEVNULL  # py3k

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
__status__ = "Development"

log = get_logger(__name__)  # pylint: disable=C0103

JOB_TEMPLATES = {
    "in2p3": lstrip(
        """
        #$ -N {job_name}
        #$ -M {email}
        ## Send mail at: start (b), completion (e), never (n)
        #$ -m {send_mail}
        #$ -j y
        #$ -o {log_path}/{job_name}{task_name}.log
        #$ -l os={platform}
        #$ -P P_{group}
        #$ -S {shell}
        {job_array_option}
        #
        ## Walltime (HH:MM:SS), for different batch systems (IN2P3, ECAP, ...)
        #$ -l ct={walltime}
        ## Memory (Units: G, M, K, B; Min 64M)
        #$ -l vmem={vmem}
        ## Local scratch diskspace (at TMPDIR)
        #$ -l fsize={fsize}
        #
        ## Extra Resources
        #$ -l irods={irods:d}
        #$ -l sps={sps:d}
        #$ -l hpss={hpss:d}
        #$ -l xrootd={xrootd:d}
        #$ -l dcache={dcache:d}
        #$ -l oracle={oracle:d}

        echo "========================================================"
        echo "Job started on" $(date)
        echo "========================================================"

        {script}

        echo "========================================================"
        echo "JAWOLLJA! Job exited on" $(date)
        echo "========================================================"
        """
    ),
    "woody": lstrip(
        """
        #PBS -N {job_name}
        #PBS -M {email} -m a
        #PBS -o {log_path}/{job_name}{task_name}.out.log
        #PBS -e {log_path}/{job_name}{task_name}.err.log
        #PBS -l walltime={walltime}

        echo "========================================================"
        echo "Job started on" $(date)
        echo "========================================================"

        {script}

        echo "========================================================"
        echo "JAWOLLJA! Job exited on" $(date)
        echo "========================================================"
        """
    ),
}


def qsub(script, job_name, dryrun=False, silent=False, *args, **kwargs):
    """Submit a job via qsub.


    Returns the job script as string.
    """
    if not silent:
        print("Preparing job script...")
    job_string = gen_job(script=script, job_name=job_name, *args, **kwargs)
    env = os.environ.copy()
    if dryrun:
        print(
            "This is a dry run! Here is the generated job file, which will "
            "not be submitted:"
        )
        print(job_string)
    else:
        if not silent:
            print("Calling qsub with the generated job script.")
            out_pipe = subprocess.PIPE
        else:
            out_pipe = DEVNULL
        p = subprocess.Popen(
            "qsub -V", stdin=subprocess.PIPE, env=env, shell=True, stdout=out_pipe
        )
        p.communicate(input=bytes(job_string.encode("ascii")))

    return job_string


def gen_job(
    script,
    job_name,
    log_path="qlogs",
    group="km3net",
    platform="cl7",
    walltime="00:10:00",
    cluster="in2p3",
    vmem="8G",
    fsize="8G",
    shell=None,
    email=None,
    send_mail="n",
    job_array_start=1,
    job_array_stop=None,
    job_array_step=1,
    irods=False,
    sps=True,
    hpss=False,
    xrootd=False,
    dcache=False,
    oracle=False,
    split_array_logs=False,
):
    """Generate a job script."""
    if shell is None:
        shell = os.environ["SHELL"]
    if email is None:
        email = os.environ["USER"] + "@km3net.de"
    if isinstance(script, Script):
        script = str(script)
    log_path = os.path.abspath(log_path)
    if job_array_stop is not None:
        job_array_option = "#$ -t {}-{}:{}".format(
            job_array_start, job_array_stop, job_array_step
        )
    else:
        job_array_option = "#"
    if split_array_logs:
        task_name = "_$TASK_ID"
    else:
        task_name = ""
    job_string = JOB_TEMPLATES[cluster].format(
        script=script,
        email=email,
        send_mail=send_mail,
        log_path=log_path,
        job_name=job_name,
        group=group,
        walltime=walltime,
        vmem=vmem,
        fsize=fsize,
        irods=irods,
        sps=sps,
        hpss=hpss,
        xrootd=xrootd,
        dcache=dcache,
        oracle=oracle,
        shell=shell,
        platform=platform,
        job_array_option=job_array_option,
        task_name=task_name,
    )
    return job_string


def get_jpp_env(jpp_dir):
    """Return the environment dict of a loaded Jpp env.

    The returned env can be passed to `subprocess.Popen("J...", env=env)`
    to execute Jpp commands.

    """
    env = {
        v[0]: "".join(v[1:])
        for v in [
            l.split("=")
            for l in os.popen("source {0}/setenv.sh {0} && env".format(jpp_dir))
            .read()
            .split("\n")
            if "=" in l
        ]
    }
    return env


class Script(object):
    """A shell script which can be built line by line for `qsub`."""

    def __init__(self):
        self.lines = []

    def add(self, line):
        """Add a new line"""
        self.lines.append(line)

    def echo(self, text):
        """Add an echo command. The given text will be double qouted."""
        self.lines.append('echo "{}"'.format(text))

    def separator(self, character="=", length=42):
        """Add a visual separator."""
        self.echo(character * length)

    def cp(self, source, target):
        """Add a new copy instruction"""
        self._add_two_argument_command("cp", source, target)

    def mv(self, source, target):
        """Add a new move instruction"""
        self._add_two_argument_command("mv", source, target)

    def mkdir(self, folder_path):
        """Add a new 'mkdir -p' instruction"""
        self.add('mkdir -p "{}"'.format(folder_path))

    def iget(self, irods_path, attempts=1, pause=15):
        """Add an iget command to retrieve a file from iRODS.

        Parameters
        ----------
            irods_path: str
                Filepath which should be fetched using iget
            attempts: int (default: 1)
                Number of retries, if iRODS access fails
            pause: int (default: 15)
                Pause between two access attempts in seconds
        """
        if attempts > 1:
            cmd = """   for i in {{1..{0}}}; do
                            ret=$(iget -v {1} 2>&1)
                            echo $ret
                            if [[ $ret == *"ERROR"* ]]; then
                                echo "Attempt $i failed"
                            else
                                break
                            fi
                            sleep {2}s
                        done """
            cmd = lstrip(cmd)
            cmd = cmd.format(attempts, irods_path, pause)
            self.add(cmd)
        else:
            self.add('iget -v "{}"'.format(irods_path))

    def _add_two_argument_command(self, command, arg1, arg2):
        """Helper function for two-argument commands"""
        self.lines.append("{} {} {}".format(command, arg1, arg2))

    def clear(self):
        self.lines = []

    def __add__(self, other):
        new_script = Script()
        new_script.lines = self.lines + other.lines
        return new_script

    def __str__(self):
        return "\n".join(self.lines)

    def __repr__(self):
        return "# Shell script\n" + str(self)



# Filename: test_controlhost.py
# pylint: disable=locally-disabled,C0111,R0904,R0201,C0103,W0612
"""
Unit tests for the controlhost module.

"""

from km3pipe.controlhost import Tag, Message, Prefix

from km3pipe.testing import TestCase

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
__status__ = "Development"


class TestTag(TestCase):
    def test_empty_tag_has_correct_length(self):
        tag = Tag()
        self.assertEqual(Tag.SIZE, len(tag))

    def test_tag_has_correct_length(self):
        for tag_name in (b"foo", b"bar", b"baz", b"1"):
            tag = Tag(tag_name)
            self.assertEqual(Tag.SIZE, len(tag))

    def test_tag_with_invalid_length_raises_valueerror(self):
        self.assertRaises(ValueError, Tag, "123456789")

    def test_tag_has_correct_data(self):
        tag = Tag(b"foo")
        self.assertEqual(b"foo\x00\x00\x00\x00\x00", tag.data)
        tag = Tag("abcdefgh")
        self.assertEqual("abcdefgh", tag.data)

    def test_tag_has_correct_string_representation(self):
        tag = Tag(b"foo")
        self.assertEqual("foo", str(tag))


class TestPrefix(TestCase):
    def test_init(self):
        Prefix(b"foo", 1)


class TestMessage(TestCase):
    def test_init(self):
        Message("")



# Filename: test_mc.py
# pylint: disable=C0111,R0904,C0103
# vim:set ts=4 sts=4 sw=4 et:
"""
Tests for MC Modules.
"""

import numpy as np
from numpy.testing import assert_array_equal, assert_allclose
import pytest

from km3pipe import Table, Blob, Pipeline, Module
from km3pipe.testing import TestCase

from km3modules.mc import (
    convert_mc_times_to_jte_times,
    MCTimeCorrector,
    GlobalRandomState,
)

__author__ = "Moritz Lotze, Michael Moser"
__copyright__ = "Copyright 2018, Tamas Gal and the KM3NeT collaboration."
__license__ = "MIT"
__maintainer__ = "Tamas Gal, Moritz Lotze"
__email__ = "tgal@km3net.de"
__status__ = "Development"


class TestMCConvert(TestCase):
    def setUp(self):
        self.event_info = Table(
            {"timestamp": 1, "nanoseconds": 700000000, "mc_time": 1.74999978e9,}
        )

        self.mc_tracks = Table({"time": 1,})

        self.mc_hits = Table({"time": 30.79,})

        self.blob = Blob(
            {
                "event_info": self.event_info,
                "mc_hits": self.mc_hits,
                "mc_tracks": self.mc_tracks,
            }
        )

    def test_convert_mc_times_to_jte_times(self):
        times_mc_tracks = convert_mc_times_to_jte_times(
            self.mc_tracks.time,
            self.event_info.timestamp * 1e9 + self.event_info.nanoseconds,
            self.event_info.mc_time,
        )
        times_mc_hits = convert_mc_times_to_jte_times(
            self.mc_hits.time,
            self.event_info.timestamp * 1e9 + self.event_info.nanoseconds,
            self.event_info.mc_time,
        )

        assert times_mc_tracks is not None
        assert times_mc_hits is not None
        print(times_mc_tracks, times_mc_hits)
        assert np.allclose(times_mc_tracks, 49999781)
        assert np.allclose(times_mc_hits, 49999810.79)

    def test_process(self):
        corr = MCTimeCorrector(
            mc_hits_key="mc_hits",
            mc_tracks_key="mc_tracks",
            event_info_key="event_info",
        )
        newblob = corr.process(self.blob)
        assert newblob["mc_hits"] is not None
        assert newblob["mc_tracks"] is not None
        assert np.allclose(newblob["mc_hits"].time, 49999810.79)
        assert np.allclose(newblob["mc_tracks"].time, 49999781)


class TestGlobalRandomState(TestCase):
    def test_default_random_state(self):
        assertAlmostEqual = self.assertAlmostEqual

        class Observer(Module):
            def configure(self):
                self.i = 0
                self.x = [0.3745401188, 0.950714306, 0.7319939418]

            def process(self, blob):
                assertAlmostEqual(self.x[self.i], np.random.rand())
                self.i += 1
                return blob

        pipe = Pipeline()
        pipe.attach(GlobalRandomState)
        pipe.attach(Observer)
        pipe.drain(3)

    def test_custom_random_state(self):
        assertAlmostEqual = self.assertAlmostEqual

        class Observer(Module):
            def configure(self):
                self.i = 0
                self.x = [0.221993171, 0.870732306, 0.206719155]

            def process(self, blob):
                assertAlmostEqual(self.x[self.i], np.random.rand())
                self.i += 1
                return blob

        pipe = Pipeline()
        pipe.attach(GlobalRandomState, seed=5)
        pipe.attach(Observer)
        pipe.drain(3)

    def test_without_pipeline_and_default_state(self):
        GlobalRandomState()
        numbers = np.arange(1, 50)
        np.random.shuffle(numbers)
        lotto_numbers = sorted(numbers[:6])
        self.assertListEqual([14, 18, 28, 45, 46, 48], lotto_numbers)

    def test_without_pipeline_with_custom_seed(self):
        GlobalRandomState(seed=23)
        numbers = np.arange(1, 50)
        np.random.shuffle(numbers)
        lotto_numbers = sorted(numbers[:6])
        self.assertListEqual([14, 15, 18, 19, 33, 44], lotto_numbers)



# Filename: cmd.py
"""
KM3Pipe command line utility.

Usage:
    km3pipe test
    km3pipe update [GIT_BRANCH]
    km3pipe createconf [--overwrite] [--dump]
    km3pipe detx DET_ID [-m] [-t T0_SET] [-c CALIBR_ID] [-o OUT]
    km3pipe detx DET_ID RUN
    km3pipe detectors [-s REGEX] [--temporary]
    km3pipe rundetsn [--temporary] RUN DETECTOR
    km3pipe retrieve DET_ID RUN [-i -o OUTFILE]
    km3pipe (-h | --help)
    km3pipe --version

Options:
    -h --help           Show this screen.
    -m                  Get the MC detector file (flips the sign of DET_ID).
    -c CALIBR_ID        Geometrical calibration ID (eg. A01466417)
    -i                  Use iRODS instead of xrootd to retrieve files.
    -o OUT              Output folder or filename.
    -t T0_SET           Time calibration ID (eg. A01466431)
    -s REGEX            Regular expression to filter the runsetup name/id.
    --temporary         Use a temporary session. [default: False]
    --overwrite         Overwrite existing config [default: False]
    DET_ID              Detector ID (eg. D_ARCA001).
    DETECTOR            Detector (eg. ARCA).
    GIT_BRANCH          Git branch to pull (eg. develop).
    RUN                 Run number.

"""
import re
import sys
import subprocess
import os
from datetime import datetime
import time

from . import version
from .tools import irods_path, xrootd_path
from .db import DBManager, we_are_in_lyon
from .hardware import Detector

from km3pipe.logger import get_logger

log = get_logger(__name__)  # pylint: disable=C0103

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal and Moritz Lotze"
__email__ = "tgal@km3net.de"
__status__ = "Development"

SPS_CACHE = "/sps/km3net/repo/data/cache"


def run_tests():
    import pytest
    import km3pipe

    pytest.main([os.path.dirname(km3pipe.__file__)])


def update_km3pipe(git_branch=""):
    if git_branch == "" or git_branch is None:
        git_branch = "master"
    if git_branch == "develop":
        import time

        log.deprecation(
            "The 'develop' branch has been moved to 'master', please "
            "use the command \n\n"
            "    km3pipe update\n\n"
            "to get the latest development version and \n\n"
            "    pip install -U km3pipe\n\n"
            "to install the latest release.\n"
            "Proceeding with the 'master' branch in 15 seconds..."
        )
        time.sleep(15)
        git_branch = "master"
    os.system(
        "pip install -U git+http://git.km3net.de/km3py/km3pipe.git@{0}".format(
            git_branch
        )
    )


def retrieve(run_id, det_id, use_irods=False, out=None):
    """Retrieve run from iRODS for a given detector (O)ID"""
    det_id = int(det_id)

    if use_irods:
        rpath = irods_path(det_id, run_id)
        cmd = "iget -Pv"
    else:
        rpath = xrootd_path(det_id, run_id)
        cmd = "xrdcp"

    filename = os.path.basename(rpath)

    if out is not None and os.path.isdir(out):
        outfile = os.path.join(out, filename)
    elif out is None:
        outfile = filename
    else:
        outfile = out

    if os.path.exists(outfile):
        print("Output file '{}' already exists.".format(outfile))
        return

    cmd += " {0} {1}".format(rpath, outfile)

    if not we_are_in_lyon():
        subprocess.call(cmd.split())
        return

    subfolder = os.path.join(*rpath.split("/")[-3:-1])
    cached_subfolder = os.path.join(SPS_CACHE, subfolder)
    cached_filepath = os.path.join(cached_subfolder, filename)
    lock_file = cached_filepath + ".in_progress"

    if os.path.exists(lock_file):
        print("File is already requested, waiting for the download to finish.")
        for _ in range(6 * 15):  # 15 minute timeout
            time.sleep(10)
            if not os.path.exists(lock_file):
                print("Done.")
                break
        else:
            print(
                "Timeout reached. Deleting the lock file and initiating a "
                "new download."
            )
            os.remove(lock_file)

    if not os.path.exists(cached_filepath):
        print("Downloading file to local SPS cache ({}).".format(SPS_CACHE))
        try:
            os.makedirs(cached_subfolder, exist_ok=True)
            subprocess.call(["touch", lock_file])
            subprocess.call(["chmod", "g+w", lock_file])
            subprocess.call(cmd.split())
            subprocess.call(["chmod", "g+w", outfile])
            subprocess.call(["cp", "-p", outfile, cached_filepath])
        except (OSError, KeyboardInterrupt) as e:
            print("\nAborting... {}".format(e))
            for f in [outfile, lock_file, cached_filepath]:
                subprocess.call(["rm", "-f", f])
            return
        finally:
            for f in [outfile, lock_file]:
                subprocess.call(["rm", "-f", f])

    subprocess.call(["ln", "-s", cached_filepath, outfile])


def detx(det_id, calibration="", t0set="", filename=None):
    now = datetime.now()
    if filename is None:
        filename = "KM3NeT_{0}{1:08d}_{2}{3}{4}.detx".format(
            "-" if det_id < 0 else "",
            abs(det_id),
            now.strftime("%d%m%Y"),
            "_t0set-%s" % t0set if t0set else "",
            "_calib-%s" % calibration if calibration else "",
        )
    det = Detector(det_id=det_id, t0set=t0set, calibration=calibration)
    if det.n_doms > 0:
        det.write(filename)


def detx_for_run(det_id, run, temporary=False):
    """Retrieve the calibrated detx for a given det_id and run"""
    db = DBManager(temporary=temporary)
    raw_detx = db.detx_for_run(det_id, run)
    filename = "KM3NeT_{0:08d}_{1:08d}.detx".format(det_id, run)
    with open(filename, "w") as fobj:
        fobj.write(raw_detx)
    print("File saved as '{}'".format(filename))


def detectors(regex=None, sep="\t", temporary=False):
    """Print the detectors table"""
    db = DBManager(temporary=temporary)
    dt = db.detectors
    if regex is not None:
        try:
            re.compile(regex)
        except re.error:
            log.error("Invalid regex!")
            return
        dt = dt[dt["OID"].str.contains(regex) | dt["CITY"].str.contains(regex)]
    dt.to_csv(sys.stdout, sep=sep)


def rundetsn(run_id, detector="ARCA", temporary=False):
    """Print the detector serial number for a given run of ARCA/ORCA"""
    db = DBManager(temporary=temporary)
    dts = db.detectors
    for det_id in dts[dts.OID.str.contains(detector)].SERIALNUMBER:
        if run_id in db.run_table(det_id).RUN.values:
            print(det_id)
            return


def createconf(overwrite=False, dump=False):
    import os.path
    from os import environ

    defaultconf = "[General]\ncheck_for_updates=yes"
    if dump:
        print("--------------\n" + defaultconf + "\n--------------")
    fname = environ["HOME"] + "/.km3net"
    if not overwrite and os.path.exists(fname):
        log.warn("Config exists, not overwriting")
        return
    with open(fname, "w") as f:
        f.write(defaultconf)
    print("Wrote default config to ", fname)
    os.chmod(fname, 0o600)


def main():
    from docopt import docopt

    args = docopt(__doc__, version="KM3Pipe {}".format(version,))

    if args["test"]:
        run_tests()

    if args["update"]:
        update_km3pipe(args["GIT_BRANCH"])

    if args["createconf"]:
        overwrite_conf = bool(args["--overwrite"])
        dump = bool(args["--dump"])
        createconf(overwrite_conf, dump)

    if args["rundetsn"]:
        rundetsn(int(args["RUN"]), args["DETECTOR"], temporary=args["--temporary"])

    if args["retrieve"]:
        retrieve(int(args["RUN"]), args["DET_ID"], use_irods=args["-i"], out=args["-o"])

    if args["detx"]:
        if args["RUN"] and args["DET_ID"]:
            det_id = int(args["DET_ID"])
            run = int(args["RUN"])
            detx_for_run(det_id, run, temporary=args["--temporary"])
        else:
            t0set = args["-t"]
            calibration = args["-c"]
            outfile = args["-o"]
            det_id = int(("-" if args["-m"] else "") + args["DET_ID"])
            detx(det_id, calibration, t0set, outfile)

    if args["detectors"]:
        detectors(regex=args["-s"], temporary=args["--temporary"])



from km3pipe import Pipeline, Module
from km3pipe.io import CHPump


class CHPrinter(Module):
    def process(self, blob):
        print("New blob:")
        print(blob["CHPrefix"])
        print(blob["CHData"])
        return blob


pipe = Pipeline()
pipe.attach(
    CHPump, host="127.0.0.1", port=5553, tags="foo, narf", timeout=1000, max_queue=42
)
pipe.attach(CHPrinter)
pipe.drain()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
==========
Provenance
==========

"""

# Author: Tamás Gál <tgal@km3net.de>
# License: MIT
# Date: 2020-08-23

#####################################################
# Introduction
# -----------
# KM3Pipe uses the provenance functionality from ``thepipe``
# which automatically tracks each activity. This document
# shows how it works.
#
import km3pipe as kp
import km3modules as km
import numpy as np

#####################################################
# Some dummy modules
# ------------------

class RandomNumberGenerator(kp.Module):
    def configure(self):
        self.h5loc = self.require("h5loc")
        self.n = self.get("n", default=10)

    def process(self, blob):
        table = kp.Table({"x": np.random.randn(self.n)}, h5loc=self.h5loc)
        blob["RandomNumbers"] = table
        return blob


#####################################################
# Creating a simple pipeline
# --------------------------
# We create a very basic pipeline:

pipe = kp.Pipeline()
pipe.attach(km.StatusBar, every=1)
pipe.attach(km.mc.GlobalRandomState, seed=23)
pipe.attach(RandomNumberGenerator, h5loc="/rnd", n=5)
pipe.attach(kp.io.HDF5Sink, filename="rnd.h5")
pipe.drain(11)


#####################################################
# Provenance
# ----------
# The provenance information is managed by the singleton class
# ``Provenance``. To access all the provenance information,
# use the ``as_json()`` method:

print(kp.Provenance().as_json(indent=2))



import numpy as np
import km3pipe as kp


class APump(kp.Pump):
    def configure(self):
        self.index = 0

    def process(self, blob):
        data = {"a": self.index * np.arange(5), "b": np.arange(5) ** self.index}
        data2 = {
            "c": self.index * np.arange(10, dtype="f4") + 0.1,
            "d": np.arange(10, dtype="f4") ** self.index + 0.2,
        }
        print(data2)
        blob["Tablelike"] = kp.Table(data, h5loc="/tablelike", name="2D Table")
        print(blob["Tablelike"])
        blob["Columnwise"] = kp.Table(
            data2, h5loc="/columnwise", split_h5=True, name="Column-wise Split"
        )
        self.index += 1
        return blob


pipe = kp.Pipeline()
pipe.attach(APump)
pipe.attach(kp.io.HDF5Sink, filename="km3hdf5_example.h5")
pipe.drain(13)



#!usr/bin/env python
# -*- coding: utf-8 -*-
# Filename: math.py
# pylint: disable=C0103
"""
Maths, Geometry, coordinates.
"""
import numpy as np

from .logger import get_logger

__author__ = "Tamas Gal and Moritz Lotze"
__copyright__ = "Copyright 2017, Tamas Gal and the KM3NeT collaboration."
__credits__ = ["Vladimir Kulikovskiy"]
__license__ = "MIT"
__maintainer__ = "Tamas Gal and Moritz Lotze"
__email__ = "tgal@km3net.de"
__status__ = "Development"

log = get_logger(__name__)  # pylint: disable=C0103


def neutrino_to_source_direction(phi, theta, radian=True):
    """Flip the direction.

    Parameters
    ==========
    phi, theta: neutrino direction
    radian: bool [default=True]
        receive + return angles in radian? (if false, use degree)
    """
    phi = np.atleast_1d(phi).copy()
    theta = np.atleast_1d(theta).copy()
    if not radian:
        phi *= np.pi / 180
        theta *= np.pi / 180
    assert np.all(phi <= 2 * np.pi)
    assert np.all(theta <= np.pi)
    azimuth = (phi + np.pi) % (2 * np.pi)
    zenith = np.pi - theta
    if not radian:
        azimuth *= 180 / np.pi
        zenith *= 180 / np.pi
    return azimuth, zenith


def source_to_neutrino_direction(azimuth, zenith, radian=True):
    """Flip the direction.

    Parameters
    ==========
    zenith, azimuth: neutrino origin
    radian: bool [default=True]
        receive + return angles in radian? (if false, use degree)
    """
    azimuth = np.atleast_1d(azimuth).copy()
    zenith = np.atleast_1d(zenith).copy()
    if not radian:
        azimuth *= np.pi / 180
        zenith *= np.pi / 180
    phi = (azimuth - np.pi) % (2 * np.pi)
    theta = np.pi - zenith
    if not radian:
        phi *= 180 / np.pi
        theta *= 180 / np.pi
    return phi, theta


def theta(v):
    """Neutrino direction in polar coordinates.

    Downgoing event: theta = 180deg
    Horizont: 90deg
    Upgoing: theta = 0

    Angles in radians.
    """
    v = np.atleast_2d(v)
    dir_z = v[:, 2]
    return theta_separg(dir_z)


def theta_separg(dir_z):
    return np.arccos(dir_z)


def phi(v):
    """Neutrino direction in polar coordinates.

    ``phi``, ``theta`` is the opposite of ``zenith``, ``azimuth``.

    Angles in radians.
    """
    v = np.atleast_2d(v)
    dir_x = v[:, 0]
    dir_y = v[:, 1]
    return phi_separg(dir_x, dir_y)


def phi_separg(dir_x, dir_y):
    p = np.arctan2(dir_y, dir_x)
    p[p < 0] += 2 * np.pi
    return p


def zenith(v):
    """Return the zenith angle in radians.

    Defined as 'Angle respective to downgoing'.
    Downgoing event: zenith = 0
    Horizont: 90deg
    Upgoing: zenith = 180deg
    """
    return angle_between((0, 0, -1), v)


def azimuth(v):
    """Return the azimuth angle in radians.

    ``phi``, ``theta`` is the opposite of ``zenith``, ``azimuth``.

    This is the 'normal' azimuth definition -- beware of how you
    define your coordinates. KM3NeT defines azimuth
    differently than e.g. SLALIB, astropy, the AAS.org
    """
    v = np.atleast_2d(v)
    azi = phi(v) - np.pi
    azi[azi < 0] += 2 * np.pi
    if len(azi) == 1:
        return azi[0]
    return azi


def cartesian(phi, theta, radius=1):
    x = radius * np.sin(theta) * np.cos(phi)
    y = radius * np.sin(theta) * np.sin(phi)
    z = radius * np.cos(theta)
    return np.column_stack((x, y, z))


def angle_between(v1, v2):
    """Returns the angle in radians between vectors 'v1' and 'v2'.

    >>> angle_between((1, 0, 0), (0, 1, 0))
    1.5707963267948966
    >>> angle_between((1, 0, 0), (1, 0, 0))
    0.0
    >>> angle_between((1, 0, 0), (-1, 0, 0))
    3.141592653589793

    """
    v1_u = unit_vector(v1)
    v2_u = unit_vector(v2)
    # Don't use `np.dot`, does not work with all shapes
    angle = np.arccos(np.inner(v1_u, v2_u))
    return angle


def innerprod_1d(v1, v2):
    """1d Inner product for vector-of-vectors.

    Example:
    ========
    ```
    v1 = np.array([dir_x, dir_y, dir_z]).T
    v2 = ... # dito
    angle_between_v1_v1 = innerprod_1d(v1, v2)
    ```
    """
    return np.einsum("ij,ij->i", v1, v1)


def unit_vector(vector, **kwargs):
    """Returns the unit vector of the vector."""
    # This also works for a dataframe with columns ['x', 'y', 'z']
    # However, the division operation is picky about the shapes
    # So, remember input vector shape, cast all up to 2d,
    # do the (ugly) conversion, then return unit in same shape as input
    # of course, the numpy-ized version of the input...
    vector = np.array(vector)
    out_shape = vector.shape
    vector = np.atleast_2d(vector)
    unit = vector / np.linalg.norm(vector, axis=1, **kwargs)[:, None]
    return unit.reshape(out_shape)


def pld3(pos, line_vertex, line_dir):
    """Calculate the point-line-distance for given point and line."""
    pos = np.atleast_2d(pos)
    line_vertex = np.atleast_1d(line_vertex)
    line_dir = np.atleast_1d(line_dir)
    c = np.cross(line_dir, line_vertex - pos)
    n1 = np.linalg.norm(c, axis=1)
    n2 = np.linalg.norm(line_dir)
    out = n1 / n2
    if out.ndim == 1 and len(out) == 1:
        return out[0]
    return out


def lpnorm(x, p=2):
    return np.power(np.sum(np.power(x, p)), 1 / p)


def dist(x1, x2, axis=0):
    """Return the distance between two points.

    Set axis=1 if x1 is a vector and x2 a matrix to get a vector of distances.
    """
    return np.linalg.norm(x2 - x1, axis=axis)


def com(points, masses=None):
    """Calculate center of mass for given points.
    If masses is not set, assume equal masses."""
    if masses is None:
        return np.average(points, axis=0)
    else:
        return np.average(points, axis=0, weights=masses)


def circ_permutation(items):
    """Calculate the circular permutation for a given list of items."""
    permutations = []
    for i in range(len(items)):
        permutations.append(items[i:] + items[:i])
    return permutations


def hsin(theta):
    """haversine"""
    return (1.0 - np.cos(theta)) / 2.0


def space_angle(phi1, theta1, phi2, theta2):
    """Also called Great-circle-distance --
    use long-ass formula from wikipedia (last in section):
    https://en.wikipedia.org/wiki/Great-circle_distance#Computational_formulas

    Space angle only makes sense in lon-lat, so convert zenith -> latitude.
    """
    from numpy import pi, sin, cos, arctan2, sqrt, square

    lamb1 = pi / 2 - theta1
    lamb2 = pi / 2 - theta2
    lambdelt = lamb2 - lamb1
    under = sin(phi1) * sin(phi2) + cos(phi1) * cos(phi2) * cos(lambdelt)
    over = sqrt(
        np.square((cos(phi2) * sin(lambdelt)))
        + square(cos(phi1) * sin(phi2) - sin(phi1) * cos(phi2) * cos(lambdelt))
    )
    angle = arctan2(over, under)
    return angle


def rotation_matrix(axis, theta):
    """The Euler–Rodrigues formula.

    Return the rotation matrix associated with counterclockwise rotation about
    the given axis by theta radians.

    Parameters
    ----------
    axis: vector to rotate around
    theta: rotation angle, in rad
    """
    axis = np.asarray(axis)
    axis = axis / np.linalg.norm(axis)
    a = np.cos(theta / 2)
    b, c, d = -axis * np.sin(theta / 2)
    aa, bb, cc, dd = a * a, b * b, c * c, d * d
    bc, ad, ac, ab, bd, cd = b * c, a * d, a * c, a * b, b * d, c * d
    return np.array(
        [
            [aa + bb - cc - dd, 2 * (bc + ad), 2 * (bd - ac)],
            [2 * (bc - ad), aa + cc - bb - dd, 2 * (cd + ab)],
            [2 * (bd + ac), 2 * (cd - ab), aa + dd - bb - cc],
        ]
    )


class Polygon(object):
    """A polygon, to implement containment conditions."""

    def __init__(self, vertices):
        from matplotlib.path import Path

        self.poly = Path(vertices)

    def contains(self, points):
        points = np.atleast_2d(points)
        points_flat = points.reshape((-1, 2))
        is_contained = self.poly.contains_points(points_flat)
        return is_contained

    def contains_xy(self, x, y):
        xy = np.column_stack((x, y))
        return self.contains(xy)


class IrregularPrism(object):
    """Like a cylinder, but the top is an irregular Polygon."""

    def __init__(self, xy_vertices, z_min, z_max):
        self.poly = Polygon(xy_vertices)
        self.z_min = z_min
        self.z_max = z_max

    def _is_z_contained(self, z):
        return (self.z_min <= z) & (z <= self.z_max)

    def contains(self, points):
        points = np.atleast_2d(points)
        points_xy = points[:, [0, 1]]
        points_z = points[:, 2]
        is_xy_contained = self.poly.contains(points_xy)
        is_z_contained = self._is_z_contained(points_z)
        return is_xy_contained & is_z_contained

    def contains_xyz(self, x, y, z):
        xyz = np.column_stack((x, y, z))
        return self.contains(xyz)


class SparseCone(object):
    """A Cone, represented by sparse samples.

    This samples evenly spaced points from the base circle.

    Parameters
    ----------
    spike_pos: coordinates of the top
    bottom_center_pos: center of the bottom circle
    opening_angle: cone opening angle, in rad
        theta, axis to mantle, *not* mantle-mantle. So this is the angle
        to the axis, and mantle-to-mantle (aperture) is 2 theta.
    """

    def __init__(self, spike_pos, bottom_center_pos, opening_angle):
        self.spike_pos = np.asarray(spike_pos)
        self.bottom_center_pos = np.asarray(bottom_center_pos)
        self.opening_angle = opening_angle
        self.top_bottom_vec = self.bottom_center_pos - self.spike_pos
        self.height = np.linalg.norm(self.top_bottom_vec)
        self.mantle_length = self.height / np.cos(self.opening_angle)
        self.radius = self.height * np.tan(self.opening_angle * self.height)

    @classmethod
    def _equidistant_angles_from_circle(cls, n_angles=4):
        return np.linspace(0, 2 * np.pi, n_angles + 1)[:-1]

    @property
    def _random_circle_vector(self):
        k = self.top_bottom_vec
        r = self.radius
        x = np.random.randn(3)
        x -= x.dot(k) * k
        x *= r / np.linalg.norm(x)
        return x

    def sample_circle(self, n_angles=4):
        angles = self._equidistant_angles_from_circle(n_angles)
        random_circle_vector = self._random_circle_vector
        # rotate the radius vector around the cone axis
        points_on_circle = [
            np.dot(rotation_matrix(self.top_bottom_vec, theta), random_circle_vector)
            for theta in angles
        ]
        return points_on_circle

    @property
    def sample_axis(self):
        return [self.spike_pos, self.bottom_center_pos]

    def sample(self, n_circle=4):
        points = self.sample_circle(n_circle)
        points.extend(self.sample_axis)
        return points


def inertia(x, y, z, weight=None):
    """Inertia tensor, stolen of thomas"""
    if weight is None:
        weight = 1
    tensor_of_inertia = np.zeros((3, 3), dtype=float)
    tensor_of_inertia[0][0] = (y * y + z * z) * weight
    tensor_of_inertia[0][1] = (-1) * x * y * weight
    tensor_of_inertia[0][2] = (-1) * x * z * weight
    tensor_of_inertia[1][0] = (-1) * x * y * weight
    tensor_of_inertia[1][1] = (x * x + z * z) * weight
    tensor_of_inertia[1][2] = (-1) * y * z * weight
    tensor_of_inertia[2][0] = (-1) * x * z * weight
    tensor_of_inertia[2][1] = (-1) * z * y * weight
    tensor_of_inertia[2][2] = (x * x + y * y) * weight

    eigen_values = np.linalg.eigvals(tensor_of_inertia)
    small_inertia = eigen_values[2][2]
    middle_inertia = eigen_values[1][1]
    big_inertia = eigen_values[0][0]
    return small_inertia, middle_inertia, big_inertia


def g_parameter(time_residual):
    """stolen from thomas"""
    mean = np.mean(time_residual)
    time_residual_prime = time_residual - np.ones(time_residual.shape) * mean
    time_residual_prime *= time_residual_prime / (-2 * 1.5 * 1.5)
    time_residual_prime = np.exp(time_residual_prime)
    g = np.sum(time_residual_prime) / len(time_residual)
    return g


def gold_parameter(time_residual):
    """stolen from thomas"""
    gold = np.exp(-1 * time_residual * time_residual / (2 * 1.5 * 1.5)) / len(
        time_residual
    )
    gold = np.sum(gold)


def log_b(arg, base):
    """Logarithm to any base"""
    return np.log(arg) / np.log(base)


def qrot(vector, quaternion):
    """Rotate a 3D vector using quaternion algebra.

    Implemented by Vladimir Kulikovskiy.

    Parameters
    ----------
    vector: np.array
    quaternion: np.array

    Returns
    -------
    np.array

    """
    t = 2 * np.cross(quaternion[1:], vector)
    v_rot = vector + quaternion[0] * t + np.cross(quaternion[1:], t)
    return v_rot


def qeuler(yaw, pitch, roll):
    """Convert Euler angle to quaternion.

    Parameters
    ----------
    yaw: number
    pitch: number
    roll: number

    Returns
    -------
    np.array

    """
    yaw = np.radians(yaw)
    pitch = np.radians(pitch)
    roll = np.radians(roll)

    cy = np.cos(yaw * 0.5)
    sy = np.sin(yaw * 0.5)
    cr = np.cos(roll * 0.5)
    sr = np.sin(roll * 0.5)
    cp = np.cos(pitch * 0.5)
    sp = np.sin(pitch * 0.5)

    q = np.array(
        (
            cy * cr * cp + sy * sr * sp,
            cy * sr * cp - sy * cr * sp,
            cy * cr * sp + sy * sr * cp,
            sy * cr * cp - cy * sr * sp,
        )
    )
    return q


def qrot_yaw(vector, heading):
    """Rotate vectors using quaternion algebra.


    Parameters
    ----------
    vector: np.array or list-like (3 elements)
    heading: the heading to rotate to [deg]

    Returns
    -------
    np.array

    """
    return qrot(vector, qeuler(heading, 0, 0))


def intersect_3d(p1, p2):
    """Find the closes point for a given set of lines in 3D.

    Parameters
    ----------
    p1 : (M, N) array_like
        Starting points
    p2 : (M, N) array_like
        End points.

    Returns
    -------
    x : (N,) ndarray
        Least-squares solution - the closest point of the intersections.

    Raises
    ------
    numpy.linalg.LinAlgError
        If computation does not converge.

    """
    v = p2 - p1
    normed_v = unit_vector(v)
    nx = normed_v[:, 0]
    ny = normed_v[:, 1]
    nz = normed_v[:, 2]
    xx = np.sum(nx ** 2 - 1)
    yy = np.sum(ny ** 2 - 1)
    zz = np.sum(nz ** 2 - 1)
    xy = np.sum(nx * ny)
    xz = np.sum(nx * nz)
    yz = np.sum(ny * nz)
    M = np.array([(xx, xy, xz), (xy, yy, yz), (xz, yz, zz)])
    x = np.sum(p1[:, 0] * (nx ** 2 - 1) + p1[:, 1] * (nx * ny) + p1[:, 2] * (nx * nz))
    y = np.sum(p1[:, 0] * (nx * ny) + p1[:, 1] * (ny * ny - 1) + p1[:, 2] * (ny * nz))
    z = np.sum(p1[:, 0] * (nx * nz) + p1[:, 1] * (ny * nz) + p1[:, 2] * (nz ** 2 - 1))
    return np.linalg.lstsq(M, np.array((x, y, z)), rcond=None)[0]



# -*- coding: utf-8 -*-
"""Read & dump events through the CH Pump.

"""

from km3pipe import Pipeline, Module
from km3pipe.io import CHPump


class CHPrinter(Module):
    def process(self, blob):
        print("New blob:")
        print(blob["CHPrefix"])
        return blob


class Dumper(Module):
    def configure(self):
        self.counter = 0

    def process(self, blob):
        if "CHData" in blob:
            tag = str(blob["CHPrefix"].tag)
            data = blob["CHData"]
            self.dump(data, tag)
        return blob

    def dump(self, data, tag):
        with open("{0}-{1:06}.dat".format(tag, self.counter), "w") as f:
            self.counter += 1
            f.write(data)


pipe = Pipeline()
pipe.attach(
    CHPump,
    host="127.0.0.1",
    port=5553,
    tags="IO_EVT, IO_TSL, IO_SUM, TRG_PARS",
    timeout=60 * 60 * 24,
    max_queue=42,
)
pipe.attach(CHPrinter)
pipe.attach(Dumper)
pipe.drain()



"""
Access the KM3NeT StreamDS DataBase service.

Usage:
    streamds
    streamds list
    streamds upload [options] CSV_FILE
    streamds info STREAM
    streamds get [-f FORMAT] STREAM [PARAMETERS...]
    streamds (-h | --help)
    streamds --version

Options:
    STREAM      Name of the stream.
    CSV_FILE    Whitespace separated data for the runsummary tables.
    PARAMETERS  List of parameters separated by space (e.g. detid=29).
    -f FORMAT   Usually 'txt' for ASCII or 'text' for UTF-8 [default: txt].
    -q          Dryrun! This will upload the parameters with a TEST_ prefix.
    -x          Do not verify the SSL certificate.
    -h --help   Show this screen.

"""

import getpass
import os
import json
import requests
import pandas as pd
import km3pipe as kp

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2017, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
__status__ = "Development"

log = kp.logger.get_logger("streamds")

RUNSUMMARY_URL = "https://km3netdbweb.in2p3.fr/jsonds/runsummarynumbers/i"
REQUIRED_COLUMNS = set(["run", "det_id", "source"])


def print_streams():
    """Print all available streams with their full description"""
    sds = kp.db.StreamDS()
    sds.print_streams()


def print_info(stream):
    """Print the information about a stream"""
    sds = kp.db.StreamDS()
    sds.help(stream)


def get_data(stream, parameters, fmt):
    """Retrieve data for given stream and parameters, or None if not found"""
    sds = kp.db.StreamDS()
    if stream not in sds.streams:
        log.error("Stream '{}' not found in the database.".format(stream))
        return
    params = {}
    if parameters:
        for parameter in parameters:
            if "=" not in parameter:
                log.error(
                    "Invalid parameter syntax '{}'\n"
                    "The correct syntax is 'parameter=value'".format(parameter)
                )
                continue
            key, value = parameter.split("=")
            params[key] = value
    data = sds.get(stream, fmt, **params)
    if data is not None:
        with pd.option_context("display.max_rows", None, "display.max_columns", None):
            print(data)
    else:
        sds.help(stream)


def available_streams():
    """Show a short list of available streams."""
    sds = kp.db.StreamDS()
    print("Available streams: ")
    print(", ".join(sorted(sds.streams)))


def upload_runsummary(csv_filename, dryrun=False, verify=False):
    """Reads the CSV file and uploads its contents to the runsummary table"""
    print("Checking '{}' for consistency.".format(csv_filename))
    if not os.path.exists(csv_filename):
        log.critical("{} -> file not found.".format(csv_filename))
        return
    try:
        df = pd.read_csv(csv_filename, delim_whitespace=True)
    except pd.errors.EmptyDataError as e:
        log.error(e)
        return

    cols = set(df.columns)

    if not REQUIRED_COLUMNS.issubset(cols):
        log.error(
            "Missing columns: {}.".format(
                ", ".join(str(c) for c in REQUIRED_COLUMNS - cols)
            )
        )
        return

    parameters = cols - REQUIRED_COLUMNS
    if len(parameters) < 1:
        log.error("No parameter columns found.")
        return

    if len(df) == 0:
        log.critical("Empty dataset.")
        return

    print(
        "Found data for parameters: {}.".format(", ".join(str(c) for c in parameters))
    )
    print("Converting CSV data into JSON")
    if dryrun:
        log.warn("Dryrun: adding 'TEST_' prefix to parameter names")
        prefix = "TEST_"
    else:
        prefix = ""
    db = kp.db.DBManager()  # noqa
    det_id_zero_mask = df["det_id"] == 0
    if sum(det_id_zero_mask) > 0:
        log.warning("Entries with 'det_id=0' found, removing them.")
        df = df[~det_id_zero_mask]
    df["det_id"] = df["det_id"].apply(db.to_det_oid)
    print(df)
    data = convert_runsummary_to_json(df, prefix=prefix)
    print("We have {:.3f} MB to upload.".format(len(data) / 1024 ** 2))

    print("Requesting database session.")
    if kp.db.we_are_in_lyon():
        session_cookie = (
            "sid=_kmcprod_134.158_lyo7783844001343100343mcprod1223user"  # noqa
        )
    else:
        session_cookie = kp.config.Config().get("DB", "session_cookie")
        if session_cookie is None:
            raise SystemExit("Could not restore DB session.")
    log.debug("Using the session cookie: {}".format(session_cookie))
    cookie_key, sid = session_cookie.split("=")
    print("Uploading the data to the database.")
    r = requests.post(
        RUNSUMMARY_URL,
        cookies={cookie_key: sid},
        files={"datafile": data},
        verify=verify,
    )
    if r.status_code == 200:
        log.debug("POST request status code: {}".format(r.status_code))
        print("Database response:")
        db_answer = json.loads(r.text)
        for key, value in db_answer.items():
            print("  -> {}: {}".format(key, value))
        if db_answer["Result"] == "OK":
            print("Upload successful.")
        else:
            log.critical("Something went wrong.")
    else:
        log.error("POST request status code: {}".format(r.status_code))
        log.critical("Something went wrong...")
        return


def convert_runsummary_to_json(
    df, comment="Uploaded via km3pipe.StreamDS", prefix="TEST_"
):
    """Convert a Pandas DataFrame with runsummary to JSON for DB upload"""
    data_field = []
    comment += ", by {}".format(getpass.getuser())
    for det_id, det_data in df.groupby("det_id"):
        runs_field = []
        data_field.append({"DetectorId": det_id, "Runs": runs_field})

        for run, run_data in det_data.groupby("run"):
            parameters_field = []
            runs_field.append({"Run": int(run), "Parameters": parameters_field})

            parameter_dict = {}
            for row in run_data.iterrows():
                for parameter_name in run_data.columns:
                    if parameter_name in REQUIRED_COLUMNS:
                        continue

                    if parameter_name not in parameter_dict:
                        entry = {"Name": prefix + parameter_name, "Data": []}
                        parameter_dict[parameter_name] = entry
                    data_value = getattr(row[1], parameter_name)
                    try:
                        data_value = float(data_value)
                    except ValueError as e:
                        log.critical("Data values has to be floats!")
                        raise ValueError(e)
                    value = {"S": str(getattr(row[1], "source")), "D": data_value}
                    parameter_dict[parameter_name]["Data"].append(value)
            for parameter_data in parameter_dict.values():
                parameters_field.append(parameter_data)
    data_to_upload = {"Comment": comment, "Data": data_field}
    file_data_to_upload = json.dumps(data_to_upload)
    return file_data_to_upload


def main():
    from docopt import docopt

    args = docopt(__doc__)

    if args["info"]:
        print_info(args["STREAM"])
    elif args["list"]:
        print_streams()
    elif args["upload"]:
        upload_runsummary(args["CSV_FILE"], args["-q"], args["-x"])
    elif args["get"]:
        get_data(args["STREAM"], args["PARAMETERS"], fmt=args["-f"])
    else:
        available_streams()



#!/usr/bin/env python
from time import ctime
from io import StringIO
from pyslack import SlackClient

from km3pipe import Pipeline, Module
from km3pipe.io import CHPump
from km3pipe.io.daq import DAQPreamble, DAQEvent


class CHPrinter(Module):
    def process(self, blob):
        print("New blob:")
        print(blob["CHPrefix"])
        return blob


class SlackSender(Module):
    def configure(self):
        self.client = SlackClient("YOUR_SLACK_API_TOKEN_HERE")
        self.current_run = None

    def process(self, blob):
        data = blob["CHData"]
        data_io = StringIO(data)

        preamble = DAQPreamble(file_obj=data_io)
        print(ctime())
        print(preamble)

        event = DAQEvent(file_obj=data_io)
        print("Run: {0}".format(event.header.run))
        print("Time Slice: {0}".format(event.header.time_slice))

        if not self.current_run == event.header.run:
            self.current_run = event.header.run
            self.client.chat_post_message(
                "#du2-live",
                "Run #{0} has started!".format(event.header.run),
                username="slackbot",
            )

        return blob


pipe = Pipeline()
pipe.attach(
    CHPump,
    host="127.0.0.1",  # You'll need an SSH tunnel
    port=5553,
    tags="IO_EVT",
    timeout=60 * 60 * 24,
    max_queue=50,
)
pipe.attach(SlackSender)
pipe.attach(CHPrinter)

pipe.drain()



# Filename: wtd.py
"""
Prints information for a given DOM (and detector [O]ID)

Usage:
    wtd DET_ID_OR_OID DOM_ID
    wtd (-h | --help)
    wtd --version

Options:
    DOM_ID          The actual DOM ID.
    DET_ID_OR_OID   Detector ID (like 29) or OID (like D_ARCA003).
    -h --help       Show this screen.

"""

import km3pipe as kp

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2018, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
__status__ = "Development"


def main():
    from docopt import docopt

    args = docopt(__doc__, version=kp.version)

    dom_id = int(args["DOM_ID"])
    det = args["DET_ID_OR_OID"]

    dom = kp.db.CLBMap(det).dom_ids[dom_id]

    print(dom)



#!/usr/bin/env python
# Filename: setup.py
"""
KM3Pipe setup script.

"""

from setuptools import setup
import sys

import builtins

# so we can detect in __init__.py that it's called from setup.py
builtins.__KM3PIPE_SETUP__ = True

with open("requirements.txt") as fobj:
    requirements = [l.strip() for l in fobj.readlines()]

with open("requirements-dev.txt") as fobj:
    requirements_dev = [l.strip() for l in fobj.readlines()]

try:
    with open("README.rst") as fh:
        long_description = fh.read()
except UnicodeDecodeError:
    long_description = "KM3Pipe"

setup(
    name="km3pipe",
    url="http://git.km3net.de/km3py/km3pipe",
    description="An analysis framework for KM3NeT",
    long_description=long_description,
    author="Tamas Gal and Moritz Lotze",
    author_email="tgal@km3net.de",
    packages=["km3pipe", "km3pipe.io", "km3pipe.utils", "km3modules", "pipeinspector"],
    include_package_data=True,
    platforms="any",
    setup_requires=["numpy>=1.12", "setuptools_scm"],
    install_requires=requirements,
    extras_require={"dev": requirements_dev},
    use_scm_version=True,
    python_requires=">=3.5",
    entry_points={
        "console_scripts": [
            "km3pipe=km3pipe.cmd:main",
            "pipeinspector=pipeinspector.app:main",
            "rtree=km3pipe.utils.rtree:main",
            "h5info=km3pipe.utils.h5info:main",
            "h5tree=km3pipe.utils.h5tree:main",
            "h5header=km3pipe.utils.h5header:main",
            "ptconcat=km3pipe.utils.ptconcat:main",
            "meantots=km3pipe.utils.meantots:main",
            "pushover=km3pipe.utils.pushover:main",
            "ztplot=km3pipe.utils.ztplot:main",
            "k40calib=km3pipe.utils.k40calib:main",
            "totmonitor=km3pipe.utils.totmonitor:main",
            "streamds=km3pipe.utils.streamds:main",
            "triggermap=km3pipe.utils.triggermap:main",
            "nb2sphx=km3pipe.utils.nb2sphx:main",
            "triggersetup=km3pipe.utils.triggersetup:main",
            "ligiermirror=km3pipe.utils.ligiermirror:main",
            "runtable=km3pipe.utils.runtable:main",
            "runinfo=km3pipe.utils.runinfo:main",
            "qrunprocessor=km3pipe.utils.qrunprocessor:main",
            "wtd=km3pipe.utils.wtd:main",
            "qrunqaqc=km3pipe.utils.qrunqaqc:main",
            "daqsample=km3pipe.utils.daqsample:main",
        ],
    },
    classifiers=[
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Programming Language :: Python",
    ],
)

__author__ = "Tamas Gal and Moritz Lotze"



# Filename: ahrs.py
# -*- coding: utf-8 -*-
# pylint: disable=locally-disabled
"""
AHRS calibration.

"""

import io
from collections import defaultdict
import time
from km3pipe.tools import timed_cache
import xml.etree.ElementTree as ET

import numpy as np
from numpy import cos, sin, arctan2
import km3pipe as kp
from km3pipe.io.daq import TMCHData

__author__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
__status__ = "Development"

log = kp.logger.get_logger(__name__)  # pylint: disable=C0103

# log.setLevel("DEBUG")


class AHRSCalibrator(kp.Module):
    """Calculates AHRS yaw, pitch and roll from median A and H of an interval.

    Required Parameters
    -------------------
    det_id: int (detector ID, e.g. 29)

    Optional Parameters
    -------------------
    interval: int (accumulation interval in [sec], default: 10s)

    Output Keys
    -----------
    'AHRSCalibration':
        dict: key=dom_id, value=tuple: (timestamp, du, floor, yaw, pitch, roll)

    """

    def configure(self):
        det_id = self.require("det_id")
        self.interval = self.get("interval") or 10  # in sec
        self.A = defaultdict(list)
        self.H = defaultdict(list)
        self.detector = kp.hardware.Detector(det_id=det_id)
        self.clbmap = kp.db.CLBMap(det_id)
        self.timestamp = time.time()

    def process(self, blob):
        tmch_data = TMCHData(io.BytesIO(blob["CHData"]))
        dom_id = tmch_data.dom_id
        try:
            du, floor, _ = self.detector.doms[dom_id]
        except KeyError:  # base CLB
            return blob

        self.A[dom_id].append(tmch_data.A)
        self.H[dom_id].append(tmch_data.H)

        if time.time() - self.timestamp > self.interval:
            self.timestamp = time.time()
            calib = self.calibrate()
            blob["AHRSCalibration"] = calib

        return blob

    def calibrate(self):
        """Calculate yaw, pitch and roll from the median of A and H.

        After successful calibration, the `self.A` and `self.H` are reset.
        DOMs with missing AHRS pre-calibration data are skipped.

        Returns
        -------
        dict: key=dom_id, value=tuple: (timestamp, du, floor, yaw, pitch, roll)

        """
        now = time.time()
        dom_ids = self.A.keys()
        print("Calibrating AHRS from median A and H for {} DOMs.".format(len(dom_ids)))
        calibrations = {}
        for dom_id in dom_ids:
            print("Calibrating DOM ID {}".format(dom_id))
            clb_upi = self.clbmap.doms_ids[dom_id].clb_upi
            ahrs_calib = get_latest_ahrs_calibration(clb_upi)
            if ahrs_calib is None:
                log.warning("AHRS calibration missing for '{}'".format(dom_id))
                continue
            du, floor, _ = self.detector.doms[dom_id]
            A = np.median(self.A[dom_id], axis=0)
            H = np.median(self.H[dom_id], axis=0)

            cyaw, cpitch, croll = fit_ahrs(A, H, *ahrs_calib)
            calibrations[dom_id] = (now, du, floor, cyaw, cpitch, croll)
        self.A = defaultdict(list)
        self.H = defaultdict(list)
        return calibrations


def fit_ahrs(A, H, Aoff, Arot, Hoff, Hrot):
    """Calculate yaw, pitch and roll for given A/H and calibration set.

    Author: Vladimir Kulikovsky

    Parameters
    ----------
    A: list, tuple or numpy.array of shape (3,)
    H: list, tuple or numpy.array of shape (3,)
    Aoff: numpy.array of shape(3,)
    Arot: numpy.array of shape(3, 3)
    Hoff: numpy.array of shape(3,)
    Hrot: numpy.array of shape(3, 3)

    Returns
    -------
    yaw, pitch, roll

    """
    Acal = np.dot(A - Aoff, Arot)
    Hcal = np.dot(H - Hoff, Hrot)

    # invert axis for DOM upside down
    for i in (1, 2):
        Acal[i] = -Acal[i]
        Hcal[i] = -Hcal[i]

    roll = arctan2(-Acal[1], -Acal[2])
    pitch = arctan2(Acal[0], np.sqrt(Acal[1] * Acal[1] + Acal[2] * Acal[2]))
    yaw = arctan2(
        Hcal[2] * sin(roll) - Hcal[1] * cos(roll),
        sum(
            (
                Hcal[0] * cos(pitch),
                Hcal[1] * sin(pitch) * sin(roll),
                Hcal[2] * sin(pitch) * cos(roll),
            )
        ),
    )

    yaw = np.degrees(yaw)
    while yaw < 0:
        yaw += 360
    # yaw = (yaw + magnetic_declination + 360 ) % 360
    roll = np.degrees(roll)
    pitch = np.degrees(pitch)
    return yaw, pitch, roll


@timed_cache(hours=1, maxsize=None, typed=False)
def get_latest_ahrs_calibration(clb_upi, max_version=3, db=None):
    """Retrieve the latest AHRS calibration data for a given CLB

    Parameters
    ----------
    clb_upi: str
    max_version: int, maximum version to check, optional
    db: DBManager(), optional

    Returns
    -------
    Aoff: numpy.array with shape(3,)
    Arot: numpy.array with shape(3,3)
    Hoff: numpy.array with shape(3,)
    Hrot: numpy.array with shape(3,3)

    or None if no calibration found.

    """
    ahrs_upi = kp.db.clbupi2ahrsupi(clb_upi)

    if db is None:
        db = kp.db.DBManager()

    datasets = []
    for version in range(max_version, 0, -1):
        for n in range(1, 100):
            log.debug("Iteration #{} to get the calib data".format(n))
            url = (
                "show_product_test.htm?upi={0}&"
                "testtype=AHRS-CALIBRATION-v{1}&n={2}&out=xml".format(
                    ahrs_upi, version, n
                )
            )
            log.debug("AHRS calib DB URL: {}".format(url))
            _raw_data = db._get_content(url).replace("\n", "")
            log.debug("What I got back as AHRS calib: {}".format(_raw_data))
            if len(_raw_data) == 0:
                break
            try:
                xroot = ET.parse(io.StringIO(_raw_data)).getroot()
            except ET.ParseError:
                continue
            else:
                datasets.append(xroot)

    if len(datasets) == 0:
        return None

    latest_dataset = _get_latest_dataset(datasets)
    return _extract_calibration(latest_dataset)


def _get_latest_dataset(datasets):
    """Find the latest valid AHRS calibration dataset"""
    return sorted(datasets, key=lambda d: d.findall(".//EndTime")[0].text)[-1]


def _extract_calibration(xroot):
    """Extract AHRS calibration information from XML root.

    Parameters
    ----------
    xroot: XML root


    Returns
    -------
    Aoff: numpy.array with shape(3,)
    Arot: numpy.array with shape(3,3)
    Hoff: numpy.array with shape(3,)
    Hrot: numpy.array with shape(3,3)

    """
    names = [c.text for c in xroot.findall(".//Name")]
    val = [[i.text for i in c] for c in xroot.findall(".//Values")]

    # The fields has to be reindeced, these are the index mappings
    col_ic = [int(v) for v in val[names.index("AHRS_Matrix_Column")]]
    try:
        row_ic = [int(v) for v in val[names.index("AHRS_Matrix_Row")]]
    except ValueError:
        row_ic = [2, 2, 2, 1, 1, 1, 0, 0, 0]
    try:
        vec_ic = [int(v) for v in val[names.index("AHRS_Vector_Index")]]
    except ValueError:
        vec_ic = [2, 1, 0]

    Aoff_ix = names.index("AHRS_Acceleration_Offset")
    Arot_ix = names.index("AHRS_Acceleration_Rotation")
    Hrot_ix = names.index("AHRS_Magnetic_Rotation")

    Aoff = np.array(val[Aoff_ix])[vec_ic].astype(float)
    Arot = (
        np.array(val[Arot_ix]).reshape(3, 3)[col_ic, row_ic].reshape(3, 3).astype(float)
    )
    Hrot = (
        np.array(val[Hrot_ix]).reshape(3, 3)[col_ic, row_ic].reshape(3, 3).astype(float)
    )

    Hoff = []
    for q in "XYZ":
        values = []
        for t in ("Min", "Max"):
            ix = names.index("AHRS_Magnetic_{}{}".format(q, t))
            values.append(float(val[ix][0]))
        Hoff.append(sum(values) / 2.0)
    Hoff = np.array(Hoff)

    return Aoff, Arot, Hoff, Hrot



# Filename: ztplot.py
"""
Create a zt-Plot.

Usage:
    ztplot [-d DETX_FILE] [-t] -e EVENT_ID FILE
    ztplot [-d DETX_FILE] [-t] -f FRAME -c COUNTER FILE
    ztplot (-h | --help)
    ztplot --version

Options:
    FILE          Input file.
    -c COUNTER    Trigger counter.
    -d DETX_FILE  Detector file.
    -e EVENT_ID   Event ID.
    -f FRAME      Frame index.
    -t            Triggered DUs only.
    -h --help     Show this screen.

"""

import matplotlib.ticker as ticker
import matplotlib.pyplot as plt
import numpy as np

import km3pipe as kp
import km3pipe.style  # noqa

km3pipe.style.use("km3pipe")

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
__status__ = "Development"


def main():
    from docopt import docopt

    arguments = docopt(__doc__)

    try:
        event_id = int(arguments["-e"])
    except TypeError:
        pass
    else:
        pump = kp.io.OfflinePump(arguments["FILE"])
        blob = pump[event_id]
        event_info = blob["EventInfo"]
        hits = blob["Hits"]
        if arguments["-d"]:
            cal = kp.calib.Calibration(filename=arguments["-d"])
        else:
            cal = kp.calib.Calibration(det_id=event_info.det_id)
        hits = cal.apply(hits)
        # triggered_dus = set(cal.detector.doms[h.dom_id][0] for h in hits)
        det = cal.detector
        if arguments["-t"]:
            dus = set(det.doms[h.dom_id][0] for h in hits if h.triggered)
        else:
            dus = set(det.doms[h.dom_id][0] for h in hits)

        n_plots = len(dus)
        n_cols = int(np.ceil(np.sqrt(n_plots)))
        n_rows = int(n_plots / n_cols) + (n_plots % n_cols > 0)
        fig, axes = plt.subplots(
            ncols=n_cols, nrows=n_rows, sharex=True, sharey=True, figsize=(16, 16)
        )

        if n_cols == 1 and n_rows == 1:
            axes = (axes,)
        else:
            axes = axes.flatten()

        for ax, du in zip(axes, dus):
            du_hits = hits[hits.du == du]
            trig_hits = du_hits.triggered_rows

            ax.scatter(
                du_hits.time - min(du_hits.time),
                du_hits.pos_z,
                c="#09A9DE",
                label="hit",
            )
            ax.scatter(
                trig_hits.time - min(du_hits.time),
                trig_hits.pos_z,
                c="#FF6363",
                label="triggered hit",
            )
            ax.set_title("DU{0}".format(du), fontsize=8, fontweight="bold")

        for ax in axes:
            ax.tick_params(labelsize=8)
            ax.yaxis.set_major_locator(ticker.MultipleLocator(50))
            xlabels = ax.get_xticklabels()
            for label in xlabels:
                label.set_rotation(45)

        plt.suptitle(
            "Filename: {0} - Event #{1}".format(arguments["FILE"], event_id),
            fontsize=16,
        )
        fig.text(0.5, 0.04, "time [ns]", ha="center")
        fig.text(0.08, 0.5, "z [m]", va="center", rotation="vertical")
        #        plt.tight_layout()
        plt.show()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
I3Pump for ANTARES DATA.
"""
import numpy as np
import km3pipe as kp
from km3pipe.logger import get_logger

from icecube import icetray, dataclasses, dataio, antares_common
from icecube.dataclasses import *

__author__ = "Nicole Geisselbrecht"
__copyright__ = "Copyright 2016, Nicole Geisselbrecht and the KM3NeT collaboration."
__credits__ = "Tamas Gal"
__license__ = "MIT"
__maintainer__ = "Nicole Geisselbrecht"
__email__ = "ngeisselbrecht@km3net.de"
__status__ = "Development"

log = get_logger(__name__)  # pylint: disable=C0103


class I3Pump(kp.Module):
    """
        Gets information of an ANTARES I3 file and converts it to HDF5.

        Attributes
        ----------
        filename : str
            Path to the input I3 file.

    """

    def configure(self):
        self.filename = self.require("filename")
        self.fobj = dataio.I3File(self.filename)
        self.blobs = self.blob_generator()
        self.omGeoMap = None
        self.cog = None

        self._hit_buffer = np.empty((10000, 14))
        self._mchit_buffer = np.empty((10000, 12))

    def process(self, blob):
        """

        Parameters
        ----------
        blob : dict
            Empty dictionary.


        Returns
        -------
        blob : dict
            Dictionary which contains information of the I3 file.

        """

        frame = next(self.blobs)

        # EVENT HEADER INFO
        if "I3EventHeader" in frame:
            header = frame["I3EventHeader"]
            blob["EventInfo"] = self._read_event_info(frame, header)

        # MC INFO
        if "AntMCTree" and "I3MCWeightDict" in frame:

            weights = kp.Table(
                {
                    "w2": self.get_antares_w2(frame),
                    "w3": self.get_antares_w3(frame),
                    "nb_gen_events": self.get_nb_generated_events(frame),
                },
                h5loc="/mc_info",
            )

            blob["McInfo"] = weights

            nu = self.get_mc_primary(frame)
            muon = self.get_mc_highest_energy_muon(frame)

            if nu is not None:
                blob["NuInfo"] = self._read_particle_info(nu, "/nu")
            if muon is not None:
                blob["MuonInfo"] = self._read_particle_info(muon, "/muon")

        # MC HITS
        if "MCHits" in frame:
            mchits = frame["MCHits"]
            blob["McHits"] = self._read_mchits(frame, mchits)

        # AAFIT
        if "AafitFinalFit" in frame:
            aafit = frame["AafitFinalFit"]
            if aafit.fit_status == dataclasses.I3Particle.FitStatus.OK:
                blob["AafitInfo"] = self._read_aafit(frame, aafit)

        # BBFIT
        if "BBFitTrack" in frame:
            bbfit_track = frame["BBFitTrack"]
            if bbfit_track.fit_status == dataclasses.I3Particle.FitStatus.OK:
                bbfit_track_chi2 = -1.0
                if "BBFitTChi2" in frame:
                    bbfit_track_chi2 = frame["BBFitTChi2"]

                blob["BbfitTrackInfo"] = self._read_bbfit(
                    bbfit_track, bbfit_track_chi2, "/reco/bbfit_track"
                )

        if "BBFitBright" in frame:
            bbfit_bright = frame["BBFitBright"]
            if bbfit_bright.fit_status == dataclasses.I3Particle.FitStatus.OK:
                bbfit_bright_chi2 = -1.0
                if "BBFitBChi2" in frame:
                    bbfit_bright_chi2 = frame["BBFitBChi2"]

                blob["BbfitBrightInfo"] = self._read_bbfit(
                    bbfit_bright, bbfit_bright_chi2, "/reco/bbfit_bright"
                )

        # GRIDFIT
        if "GridFit_FinalFitResult" in frame:
            gridfit = frame["GridFit_FinalFitResult"]
            if gridfit.fit_status == dataclasses.I3Particle.FitStatus.OK:
                blob["GridFitInfo"] = self._read_gridfit(gridfit, "/reco/gridfit")

        # HIT INFO
        if "CalibratedPulses" in frame:
            pulses = frame["CalibratedPulses"]
            blob["HitsInfo"] = self._read_hits(frame, pulses)

        return blob

    @staticmethod
    def _read_event_info(frame, header):
        """
        Reads event information of the I3 file and returns it as dict.
        """
        triggercounter = -1
        if "AntaresTriggerCounter" in frame:
            trig = frame["AntaresTriggerCounter"]
            triggercounter = trig.value

        triggermask = 0
        if "AntTriggerHierarchy" in frame:
            trigger_types = {
                "ANT_TRIGGER_3D": 1,
                "ANT_TRIGGER_1D": 2,
                "ANT_TRIGGER_1D_WITH_PREFIT": 3,
                "ANT_TRIGGER_1D_MIXED": 4,
                "ANT_TRIGGER_1D_MIXED_WITH_PREFIT": 5,
                "ANT_TRIGGER_3D_SCAN": 6,
                "ANT_TRIGGER_3D_RECURSIVE": 7,
                "ANT_TRIGGER_OB": 8,
                "ANT_TRIGGER_MINIMUM_BIAS": 9,
                "ANT_TRIGGER_1S": 10,
                "ANT_TRIGGER_3S": 11,
                "ANT_TRIGGER_3S_SCAN": 12,
                "ANT_TRIGGER_TIMESTAMP_0": 13,
                "ANT_TRIGGER_T3": 14,
                "ANT_TRIGGER_T2": 15,
                "ANT_TRIGGER_TQ": 16,
                "ANT_TRIGGER_GC": 30,
                "ANT_MERGED": 32,
            }
            tree = frame["AntTriggerHierarchy"]
            for item in tree:
                trig_type = str(item.Key.Type)
                if trig_type in trigger_types:
                    triggermask = triggermask + 2 ** (trigger_types[trig_type] - 1)

        event_info = kp.Table(
            {
                "runID": header.RunID,
                "eventID": header.EventID,
                "triggercounter": triggercounter,
                "startTimeUnix": str(header.StartTime.GetUnixTime()),
                "startTime": str(header.StartTime),
                "triggermask": triggermask,
            },
            h5loc="/event_info",
        )
        return event_info

    @staticmethod
    def _read_particle_info(particle, h5loc):
        """
        Reads particle information of the I3 file and returns it as dict.

        """
        particle_mc = kp.Table(
            {
                "dir_x": particle.GetDir().GetX(),
                "dir_y": particle.GetDir().GetY(),
                "dir_z": particle.GetDir().GetZ(),
                "pos_x": particle.GetPos().X,
                "pos_y": particle.GetPos().Y,
                "pos_z": particle.GetPos().Z,
                "energy": particle.GetEnergy(),
                "pdgencoding": particle.GetPdgEncoding(),
            },
            h5loc=h5loc,
        )
        return particle_mc

    @staticmethod
    def _read_aafit(frame, aafit):
        """
        Reads aafit information of the I3 file and returns it as dict.
        """
        lambda_value = -1
        beta_value = -1
        if "AafitLambdaFinalFit" in frame:
            lambda_frame = frame["AafitLambdaFinalFit"]
            lambda_value = lambda_frame.value

        if "AafitErrorEstimateFinalFit" in frame:
            beta_frame = frame["AafitErrorEstimateFinalFit"]
            beta_value = beta_frame.value

        aafit_pos = aafit.GetPos()
        aafit_dir = aafit.GetDir()

        aafit_info = kp.Table(
            {
                "dir_x": aafit_dir.GetX(),
                "dir_y": aafit_dir.GetY(),
                "dir_z": aafit_dir.GetZ(),
                "pos_x": aafit_pos.X,
                "pos_y": aafit_pos.Y,
                "pos_z": aafit_pos.Z,
                "lambda": lambda_value,
                "beta": beta_value,
            },
            h5loc="/reco/aafit",
        )
        return aafit_info

    @staticmethod
    def _read_bbfit(bbfit, bbfit_chi2, h5loc):
        """
        Reads bbfit information of the I3 file and returns it as dict.
        """
        bbfit_pos = bbfit.GetPos()
        bbfit_dir = bbfit.GetDir()

        bbfit_info = kp.Table(
            {
                "dir_x": bbfit_dir.GetX(),
                "dir_y": bbfit_dir.GetY(),
                "dir_z": bbfit_dir.GetZ(),
                "pos_x": bbfit_pos.X,
                "pos_y": bbfit_pos.Y,
                "pos_z": bbfit_pos.Z,
                "chi2": bbfit_chi2.value,
            },
            h5loc=h5loc,
        )
        return bbfit_info

    @staticmethod
    def _read_gridfit(gridfit, h5loc):
        """
        Reads gridfit information of the I3 file and returns it as dict.
        """
        gridfit_pos = gridfit.GetPos()
        gridfit_dir = gridfit.GetDir()
        gridfit_info = kp.Table(
            {
                "dir_x": gridfit_dir.GetX(),
                "dir_y": gridfit_dir.GetY(),
                "dir_z": gridfit_dir.GetZ(),
                "pos_x": gridfit_pos.X,
                "pos_y": gridfit_pos.Y,
                "pos_z": gridfit_pos.Z,
                "lik": gridfit.GetLik(),
            },
            h5loc=h5loc,
        )
        return gridfit_info

    def _read_hits(self, frame, pulses):
        """Creates a hits Table from calibrated pulses"""
        idx = 0
        if pulses is not None:
            for omkey, hitseries in pulses:
                omgeo = self.omGeoMap[omkey]
                rate = self.get_rate(frame, omkey)
                for hit in hitseries:

                    if idx == len(self._hit_buffer):
                        self.increase_hits_buffer_size()
                    floor = 1 + (omkey.om - 1) / 3
                    pmt = (omkey.om + 2) % 3

                    self._hit_buffer[idx][0] = hit.ID
                    self._hit_buffer[idx][1] = omkey.string
                    self._hit_buffer[idx][2] = floor
                    self._hit_buffer[idx][3] = pmt
                    self._hit_buffer[idx][4] = omgeo.position.X - self.cog.X
                    self._hit_buffer[idx][5] = omgeo.position.Y - self.cog.Y
                    self._hit_buffer[idx][6] = omgeo.position.Z - self.cog.Z
                    self._hit_buffer[idx][7] = omgeo.orientation.GetDirX()
                    self._hit_buffer[idx][8] = omgeo.orientation.GetDirY()
                    self._hit_buffer[idx][9] = omgeo.orientation.GetDirZ()
                    self._hit_buffer[idx][10] = hit.Time
                    self._hit_buffer[idx][11] = hit.Charge
                    self._hit_buffer[idx][12] = rate
                    self._hit_buffer[idx][13] = hit.IsTriggered()
                    idx += 1

            buf = self._hit_buffer[:idx]

            hits = kp.Table(
                {
                    "id": buf[:, 0].astype(int),
                    "line": buf[:, 1].astype(np.uint8),
                    "floor": buf[:, 2].astype(np.uint8),
                    "pmt": buf[:, 3].astype(np.uint32),
                    "x": buf[:, 4],
                    "y": buf[:, 5],
                    "z": buf[:, 6],
                    "dx": buf[:, 7],
                    "dy": buf[:, 8],
                    "dz": buf[:, 9],
                    "time": buf[:, 10].astype(np.float32),
                    "charge": buf[:, 11],
                    "rate": buf[:, 12],
                    "triggered": buf[:, 13],
                },
                h5loc="/hits",
                split_h5=True,
            )
            return hits

    def increase_hits_buffer_size(self):
        """
        Increases the size of the hit buffer.
        """
        current_buffer_size = len(self._hit_buffer)
        new_buffer_size = int(current_buffer_size * 1.3)
        new_buffer = np.empty((new_buffer_size, 14))
        new_buffer[:current_buffer_size] = self._hit_buffer
        self._hit_buffer = new_buffer

    def _read_mchits(self, frame, mchits):
        """Creates a mchits Table from MCHits"""
        idx = 0
        if mchits is not None:
            for omkey, hitseries in mchits:
                om_geo = self.omGeoMap[omkey]
                rate = self.get_rate(frame, omkey)
                for hit in hitseries:
                    if idx == len(self._mchit_buffer):
                        self.increase_mchits_buffer_size()
                    floor = 1 + (omkey.om - 1) / 3
                    pmt = (omkey.om + 2) % 3
                    self._mchit_buffer[idx][0] = hit.HitID
                    self._mchit_buffer[idx][1] = omkey.string
                    self._mchit_buffer[idx][2] = floor
                    self._mchit_buffer[idx][3] = pmt
                    self._mchit_buffer[idx][4] = om_geo.position.X - self.cog.X
                    self._mchit_buffer[idx][5] = om_geo.position.Y - self.cog.Y
                    self._mchit_buffer[idx][6] = om_geo.position.Z - self.cog.Z
                    self._mchit_buffer[idx][7] = om_geo.orientation.GetDirX()
                    self._mchit_buffer[idx][8] = om_geo.orientation.GetDirY()
                    self._mchit_buffer[idx][9] = om_geo.orientation.GetDirZ()
                    self._mchit_buffer[idx][10] = hit.Time
                    self._mchit_buffer[idx][11] = rate
                    idx += 1

            mcbuf = self._mchit_buffer[:idx]

            mchits = kp.Table(
                {
                    "id": mcbuf[:, 0].astype(int),
                    "line": mcbuf[:, 1].astype(np.uint8),
                    "floor": mcbuf[:, 2].astype(np.uint8),
                    "pmt": mcbuf[:, 3].astype(np.uint32),
                    "x": mcbuf[:, 4],
                    "y": mcbuf[:, 5],
                    "z": mcbuf[:, 6],
                    "dx": mcbuf[:, 7],
                    "dy": mcbuf[:, 8],
                    "dz": mcbuf[:, 9],
                    "time": mcbuf[:, 10].astype(np.float32),
                    "rate": mcbuf[:, 11],
                },
                h5loc="/mchits",
                split_h5=True,
            )
            return mchits

    def increase_mchits_buffer_size(self):
        """
        Increases the size of the mchit buffer.
        """
        current_mcbuffer_size = len(self._mchit_buffer)
        new_mcbuffer_size = int(current_mcbuffer_size * 1.3)
        new_mcbuffer = np.empty((new_mcbuffer_size, 12))
        new_mcbuffer[:current_mcbuffer_size] = self._mchit_buffer
        self._mchit_buffer = new_mcbuffer

    @staticmethod
    def get_mc_highest_energy_muon(frame):
        mctree = frame.Get("AntMCTree")
        mctracksinice = mctree.GetInIce()
        mctrack = None
        highest_energy = -1.0

        for track in mctracksinice:
            if (track.GetType() == dataclasses.I3Particle.ParticleType.MuPlus) or (
                track.GetType() == dataclasses.I3Particle.ParticleType.MuMinus
            ):
                if track.GetEnergy() > highest_energy:
                    highest_energy = track.GetEnergy()
                    mctrack = track

        return mctrack

    @staticmethod
    def get_mc_primary(frame):
        mctree = frame.Get("AntMCTree")
        primaries = mctree.GetPrimaries()
        p = None
        highest_energy = -1.0

        for primary in primaries:
            if primary.GetEnergy() > highest_energy:
                highest_energy = primary.GetEnergy()
                p = primary

        return p

    @staticmethod
    def get_antares_w3(frame):
        """
        Gets weight w3 for event.
        """
        kgen_hen_year = 3.156e7
        weight_dict = frame.Get("I3MCWeightDict")
        w3 = weight_dict["GlobalWeight"] * (
            kgen_hen_year * icetray.I3Units.second
        )  # [1/yr]
        return w3

    @staticmethod
    def get_antares_w2(frame):
        """
        Gets weight w2 for event.
        """
        kgen_hen_year = 3.156e7
        weight_dict = frame.Get("I3MCWeightDict")
        # from OneWeight in units: [GeV.cm^2.s.sr]
        # to W2 in Antares units:  [GeV.m^2.(s/yr).sr]
        w2 = (
            weight_dict["OneWeight"]
            * (icetray.I3Units.centimeter2 / icetray.I3Units.meter2)
            * kgen_hen_year
        )
        return w2

    @staticmethod
    def get_nb_generated_events(frame):
        weight_dict = frame.Get("I3MCWeightDict")
        return weight_dict["NEvents"]

    @staticmethod
    def get_rate(frame, omkey):

        rate = -1.0

        if "OMCondition" in frame:
            condition_map = frame["OMCondition"]
            condition = condition_map[omkey]
            rate = condition.GetRate() / icetray.I3Units.kilohertz

        return rate

    def blob_generator(self):
        """
        Loop over frames in I3 file.
        Reads out geometry information of geometry frames and stores it.
        Yields physics frames.
        """
        for frame in self.fobj:

            # GEOMETRY
            if frame.Stop == icetray.I3Frame.Geometry:
                geo = frame["I3Geometry"]
                self.omGeoMap = geo.omgeo
                self.cog = geo.GetDetectorCOG()

            # PHYSICS
            elif frame.Stop == icetray.I3Frame.Physics:
                yield frame

    def finish(self):
        print("finished")



# Filename: test_cmd.py
# pylint: disable=C0111,E1003,R0904,C0103,R0201,C0102

from km3pipe.testing import TestCase, patch
from km3pipe.cmd import detx, update_km3pipe

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
__status__ = "Development"

KM3PIPE_GIT = "http://git.km3net.de/km3py/km3pipe.git"


class TestDetx(TestCase):
    @patch("km3pipe.cmd.Detector")
    def test_detector_called_with_correct_det_id(self, mock_detector):
        det = mock_detector.return_value
        det.n_doms = 0
        detx(1)
        mock_detector.assert_called_with(t0set="", det_id=1, calibration="")

    @patch("km3pipe.cmd.Detector")
    def test_detector_write_called_with_correct_filename(self, mock_detector):
        det = mock_detector.return_value
        det.n_doms = 1
        detx(1)
        self.assertTrue(det.write.call_args[0][0].startswith("KM3NeT_00000001_"))
        self.assertTrue(det.write.call_args[0][0].endswith(".detx"))

    @patch("km3pipe.cmd.Detector")
    def test_detector_called_with_correct_args(self, mock_detector):
        det = mock_detector.return_value
        det.n_doms = 1
        detx(1, 2, 3)
        mock_detector.assert_called_with(t0set=3, det_id=1, calibration=2)


class TestUpdateKm3pipe(TestCase):
    @patch("km3pipe.cmd.os")
    def test_update_without_args_updates_master(self, mock_os):
        update_km3pipe()
        expected = "pip install -U git+{0}@master".format(KM3PIPE_GIT)
        mock_os.system.assert_called_with(expected)
        update_km3pipe("")
        expected = "pip install -U git+{0}@master".format(KM3PIPE_GIT)
        mock_os.system.assert_called_with(expected)
        update_km3pipe(None)
        expected = "pip install -U git+{0}@master".format(KM3PIPE_GIT)
        mock_os.system.assert_called_with(expected)

    @patch("km3pipe.cmd.os")
    def test_update_branch(self, mock_os):
        branch = "foo"
        update_km3pipe(branch)
        expected = "pip install -U git+{0}@{1}".format(KM3PIPE_GIT, branch)
        mock_os.system.assert_called_with(expected)



# Filename: hardware.py
# pylint: disable=locally-disabled
"""
Classes representing KM3NeT hardware.

"""
from collections import OrderedDict, defaultdict
from io import StringIO

import numpy as np

from .dataclasses import Table
from .tools import unpack_nfirst, split
from .math import intersect_3d, qrot_yaw  # , ignored
from .db import DBManager

from .logger import get_logger, get_printer

log = get_logger(__name__)  # pylint: disable=C0103

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
__status__ = "Development"


class Detector(object):
    """A KM3NeT detector.

    Parameters
    ----------
    filename: str, optional
        Name of .detx file with detector definition.
    det_id: int, optional
        .detx ID of detector (when retrieving from database).
    t0set: optional
        t0set (when retrieving from database).
    calibration: optional
        calibration (when retrieving from database).
    """

    def __init__(
        self, filename=None, det_id=None, t0set=None, calibration=None, string=None
    ):
        self._det_file = None
        self.det_id = None
        self.n_doms = None
        self.dus = []
        self.n_pmts_per_dom = None
        self.doms = OrderedDict()
        self.pmts = []
        self.version = None
        self.valid_from = None
        self.valid_until = None
        self.utm_info = None
        self._comments = []
        self._dom_ids = []
        self._pmt_index_by_omkey = OrderedDict()
        self._pmt_index_by_pmt_id = OrderedDict()
        self._current_du = None
        self._com = None

        self._dom_positions = None
        self._dom_table = None
        self._pmt_angles = None
        self._xy_positions = None
        self.reset_caches()

        self.cprint = get_printer(self.__class__.__name__)

        if string:
            self._init_from_string(string)

        if filename:
            self._init_from_file(filename)

        if det_id is not None:
            self.cprint(
                "Retrieving DETX with detector ID {0} "
                "from the database...".format(det_id)
            )
            db = DBManager()
            detx = db.detx(det_id, t0set=t0set, calibration=calibration)
            self._det_file = StringIO(detx)
            self._parse_header()
            self._parse_doms()
            if self.n_doms < 1:
                log.error("No data found for detector ID %s." % det_id)

    def _init_from_string(self, string):
        # TODO this is ugly, refactor me please
        self._det_file = StringIO(string)
        self._extract_comments()
        self._parse_header()
        self._parse_doms()
        self._det_file.close()

    def _init_from_file(self, filename):
        """Create detector from detx file."""
        if not filename.endswith("detx"):
            raise NotImplementedError("Only the detx format is supported.")
        self._open_file(filename)
        self._extract_comments()
        self._parse_header()
        self._parse_doms()
        self._det_file.close()

    def _open_file(self, filename):
        """Create the file handler"""
        self._det_file = open(filename, "r")

    def _readline(self, ignore_comments=True):
        """The next line of the DETX file, optionally ignores comments"""
        while True:
            line = self._det_file.readline()
            if line == "":
                return line  # To conform the EOF behaviour of .readline()
            line = line.strip()
            if line == "":
                continue  # white-space-only line
            if line.startswith("#"):
                if not ignore_comments:
                    return line
            else:
                return line

    def _extract_comments(self):
        """Retrieve all comments from the file"""
        self._det_file.seek(0, 0)
        for line in self._det_file.readlines():
            line = line.strip()
            if line.startswith("#"):
                self.add_comment(line[1:])

    def _parse_header(self):
        """Extract information from the header of the detector file"""
        self.cprint("Parsing the DETX header")
        self._det_file.seek(0, 0)
        first_line = self._readline()
        try:
            self.det_id, self.n_doms = split(first_line, int)
            self.version = "v1"
        except ValueError:
            det_id, self.version = first_line.split()
            self.det_id = int(det_id)
            self.version = self.version.lower()
            validity = self._readline().strip()
            self.valid_from, self.valid_until = split(validity, float)
            raw_utm_info = self._readline().strip().split()
            try:
                self.utm_info = UTMInfo(*raw_utm_info[1:])
            except TypeError:
                log.warning("Missing UTM information.")
            n_doms = self._readline()
            self.n_doms = int(n_doms)

    # pylint: disable=C0103
    def _parse_doms(self):
        """Extract dom information from detector file"""
        self.cprint("Reading PMT information...")
        self._det_file.seek(0, 0)
        self._readline()
        pmts = defaultdict(list)
        pmt_index = 0
        while True:
            line = self._readline()

            if line == "":
                self.cprint("Done.")
                break

            try:
                dom_id, du, floor, n_pmts = split(line, int)
            except ValueError:
                continue

            if du != self._current_du:
                log.debug("Next DU, resetting floor to 1.")
                self._current_du = du
                self.dus.append(du)
                self._current_floor = 1
                if du == 1 and floor == -1:
                    log.warning(
                        "Floor ID is -1 (Jpp conversion bug), "
                        "using our own floor ID!"
                    )
            else:
                self._current_floor += 1

            if floor == -1:
                log.debug("Setting floor ID to our own ID")
                floor = self._current_floor

            self.doms[dom_id] = (du, floor, n_pmts)

            if self.n_pmts_per_dom is None:
                self.n_pmts_per_dom = n_pmts

            if self.n_pmts_per_dom != n_pmts:
                log.warning(
                    "DOMs with different number of PMTs are "
                    "detected, this can cause some unexpected "
                    "behaviour."
                )

            for i in range(n_pmts):
                raw_pmt_info = self._readline()
                pmt_info = raw_pmt_info.split()
                pmt_id, x, y, z, rest = unpack_nfirst(pmt_info, 4)
                dx, dy, dz, t0, rest = unpack_nfirst(rest, 4)
                pmt_id = int(pmt_id)
                omkey = (du, floor, i)
                pmts["pmt_id"].append(int(pmt_id))
                pmts["pos_x"].append(float(x))
                pmts["pos_y"].append(float(y))
                pmts["pos_z"].append(float(z))
                pmts["dir_x"].append(float(dx))
                pmts["dir_y"].append(float(dy))
                pmts["dir_z"].append(float(dz))
                pmts["t0"].append(float(t0))
                pmts["du"].append(int(du))
                pmts["floor"].append(int(floor))
                pmts["channel_id"].append(int(i))
                pmts["dom_id"].append(int(dom_id))
                if self.version == "v3" and rest:
                    status, rest = unpack_nfirst(rest, 1)
                    pmts["status"].append(int(status))
                if rest:
                    log.warning("Unexpected PMT values: {0}".format(rest))
                self._pmt_index_by_omkey[omkey] = pmt_index
                self._pmt_index_by_pmt_id[pmt_id] = pmt_index
                pmt_index += 1

        self.pmts = Table(pmts, name="PMT")

    def reset_caches(self):
        log.debug("Resetting caches.")
        self._dom_positions = OrderedDict()
        self._dom_table = None
        self._xy_positions = []
        self._pmt_angles = []
        self._com = None

    def add_comment(self, comment):
        """Add a comment which will be prefixed with a '#'"""
        self._comments.append(comment)

    @property
    def comments(self):
        return self._comments

    @property
    def dom_ids(self):
        if not self._dom_ids:
            self._dom_ids = list(self.doms.keys())
        return self._dom_ids

    @property
    def dom_positions(self):
        """The positions of the DOMs, calculated from PMT directions."""
        if not self._dom_positions:
            for dom_id in self.dom_ids:
                mask = self.pmts.dom_id == dom_id
                pmt_pos = self.pmts[mask].pos
                pmt_dir = self.pmts[mask].dir
                centre = intersect_3d(pmt_pos, pmt_pos - pmt_dir * 10)
                self._dom_positions[dom_id] = centre
        return self._dom_positions

    @property
    def dom_table(self):
        """A `Table` containing DOM attributes"""
        if self._dom_table is None:
            data = defaultdict(list)
            for dom_id, (du, floor, _) in self.doms.items():
                data["dom_id"].append(dom_id)
                data["du"].append(du)
                data["floor"].append(floor)
                dom_position = self.dom_positions[dom_id]
                data["pos_x"].append(dom_position[0])
                data["pos_y"].append(dom_position[1])
                data["pos_z"].append(dom_position[2])
            self._dom_table = Table(data, name="DOMs", h5loc="/dom_table")
        return self._dom_table

    @property
    def com(self):
        """Center of mass, calculated from the mean of the PMT positions"""
        if self._com is None:
            self._com = np.mean(self.pmts.pos, axis=0)
        return self._com

    @property
    def xy_positions(self):
        """XY positions of the DUs, given by the DOMs on floor 1."""
        if self._xy_positions is None or len(self._xy_positions) == 0:
            xy_pos = []
            for dom_id, pos in self.dom_positions.items():
                if self.domid2floor(dom_id) == 1:
                    xy_pos.append(np.array([pos[0], pos[1]]))
            self._xy_positions = np.array(xy_pos)
        return self._xy_positions

    def translate_detector(self, vector):
        """Translate the detector by a given vector"""
        vector = np.array(vector, dtype=float)
        self.pmts.pos_x += vector[0]
        self.pmts.pos_y += vector[1]
        self.pmts.pos_z += vector[2]
        self.reset_caches()

    def rotate_dom_by_yaw(self, dom_id, heading, centre_point=None):
        """Rotate a DOM by a given (yaw) heading."""
        pmts = self.pmts[self.pmts.dom_id == dom_id]
        if centre_point is None:
            centre_point = self.dom_positions[dom_id]

        for pmt in pmts:
            pmt_pos = np.array([pmt.pos_x, pmt.pos_y, pmt.pos_z])
            pmt_dir = np.array([pmt.dir_x, pmt.dir_y, pmt.dir_z])
            pmt_radius = np.linalg.norm(centre_point - pmt_pos)
            index = self._pmt_index_by_pmt_id[pmt.pmt_id]
            pmt_ref = self.pmts[index]

            dir_rot = qrot_yaw([pmt.dir_x, pmt.dir_y, pmt.dir_z], heading)
            pos_rot = pmt_pos - pmt_dir * pmt_radius + dir_rot * pmt_radius

            pmt_ref.dir_x = dir_rot[0]
            pmt_ref.dir_y = dir_rot[1]
            pmt_ref.dir_z = dir_rot[2]
            pmt_ref.pos_x = pos_rot[0]
            pmt_ref.pos_y = pos_rot[1]
            pmt_ref.pos_z = pos_rot[2]
        self.reset_caches()

    def rotate_du_by_yaw(self, du, heading):
        """Rotate all DOMs on DU by a given (yaw) heading."""
        mask = self.pmts.du == du
        dom_ids = np.unique(self.pmts.dom_id[mask])
        for dom_id in dom_ids:
            self.rotate_dom_by_yaw(dom_id, heading)
        self.reset_caches()

    def rescale(self, factor, origin=(0, 0, 0)):
        """Stretch or shrink detector (DOM positions) by a given factor."""
        pmts = self.pmts
        for dom_id in self.dom_ids:
            mask = pmts.dom_id == dom_id
            pos_x = pmts[mask].pos_x
            pos_y = pmts[mask].pos_y
            pos_z = pmts[mask].pos_z
            pmts.pos_x[mask] = (pos_x - origin[0]) * factor
            pmts.pos_y[mask] = (pos_y - origin[1]) * factor
            pmts.pos_z[mask] = (pos_z - origin[2]) * factor
        self.reset_caches()

    @property
    def pmt_angles(self):
        """A list of PMT directions sorted by PMT channel, on DU-1, floor-1"""
        if self._pmt_angles == []:
            mask = (self.pmts.du == 1) & (self.pmts.floor == 1)
            self._pmt_angles = self.pmts.dir[mask]
        return self._pmt_angles

    @property
    def ascii(self):
        """The ascii representation of the detector"""
        comments = ""
        if self.version == "v3":
            for comment in self.comments:
                if not comment.startswith(" "):
                    comment = " " + comment
                comments += "#" + comment + "\n"

        if self.version == "v1":
            header = "{det.det_id} {det.n_doms}".format(det=self)
        else:
            header = "{det.det_id} {det.version}".format(det=self)
            header += "\n{0} {1}".format(self.valid_from, self.valid_until)
            header += "\n" + str(self.utm_info) + "\n"
            header += str(self.n_doms)

        doms = ""
        for dom_id, (line, floor, n_pmts) in self.doms.items():
            doms += "{0} {1} {2} {3}\n".format(dom_id, line, floor, n_pmts)
            for channel_id in range(n_pmts):
                pmt_idx = self._pmt_index_by_omkey[(line, floor, channel_id)]
                pmt = self.pmts[pmt_idx]
                doms += " {0} {1} {2} {3} {4} {5} {6} {7}".format(
                    pmt.pmt_id,
                    pmt.pos_x,
                    pmt.pos_y,
                    pmt.pos_z,
                    pmt.dir_x,
                    pmt.dir_y,
                    pmt.dir_z,
                    pmt.t0,
                )
                if self.version == "v3":
                    doms += " {0}".format(pmt.status)
                doms += "\n"
        return comments + header + "\n" + doms

    def write(self, filename):
        """Save detx file."""
        with open(filename, "w") as f:
            f.write(self.ascii)
        self.cprint("Detector file saved as '{0}'".format(filename))

    def pmt_with_id(self, pmt_id):
        """Get PMT with global pmt_id"""
        try:
            return self.pmts[self._pmt_index_by_pmt_id[pmt_id]]
        except KeyError:
            raise KeyError("No PMT found for ID: {0}".format(pmt_id))

    def get_pmt(self, dom_id, channel_id):
        """Return PMT with DOM ID and DAQ channel ID"""
        du, floor, _ = self.doms[dom_id]
        pmt = self.pmts[self._pmt_index_by_omkey[(du, floor, channel_id)]]
        return pmt

    def pmtid2omkey(self, pmt_id):
        return self._pmts_by_id[int(pmt_id)].omkey

    def domid2floor(self, dom_id):
        _, floor, _ = self.doms[dom_id]
        return floor

    @property
    def n_dus(self):
        return len(self.dus)

    def __str__(self):
        return "Detector id: '{0}', n_doms: {1}, dus: {2}".format(
            self.det_id, self.n_doms, self.dus
        )

    def __repr__(self):
        return self.__str__()


class UTMInfo(object):
    """The UTM information"""

    def __init__(self, ellipsoid, grid, easting, northing, z):
        self.ellipsoid = ellipsoid
        self.grid = grid
        self.easting = float(easting)
        self.northing = float(northing)
        self.z = float(z)

    def __str__(self):
        return "UTM {} {} {} {} {}".format(
            self.ellipsoid, self.grid, self.easting, self.northing, self.z
        )

    def __repr__(self):
        return "UTMInfo: {}".format(self)


class PMT(object):
    """Represents a photomultiplier.

    Parameters
    ----------
    id: int
    pos: 3-float-tuple (x, y, z)
    dir: 3-float-tuple (x, y, z)
    t0: int
    channel_id: int
    omkey: int
    """

    def __init__(self, id, pos, dir, t0, channel_id, omkey):
        self.id = id
        self.pos = pos
        self.dir = dir
        self.t0 = t0
        self.channel_id = channel_id
        self.omkey = omkey

    def __str__(self):
        return "PMT id:{0}, pos: {1}, dir: dir{2}, t0: {3}, DAQ channel: {4}".format(
            self.id, self.pos, self.dir, self.t0, self.channel_id
        )


# PMT DAQ channel IDs ordered from top to bottom
ORDERED_PMT_IDS = [
    28,
    23,
    22,
    21,
    27,
    29,
    20,
    30,
    26,
    25,
    19,
    24,
    13,
    7,
    1,
    14,
    18,
    12,
    6,
    2,
    11,
    8,
    0,
    15,
    4,
    3,
    5,
    17,
    10,
    9,
    16,
]



# -*- coding: utf-8 -*-
"""

==================
ToT histogram.
==================

Create a simple histogram of the PMT signals (ToTs) in all events.
"""

# Author: Tamas Gal <tgal@km3net.de>
# License: BSD-3

import tables as tb
import matplotlib.pyplot as plt
import km3pipe.style

km3pipe.style.use("km3pipe")

filename = "data/atmospheric_muons_sample.h5"

with tb.File(filename) as f:
    tots = f.get_node("/hits/tot")[:]

plt.hist(tots, bins=max(tots), log=True, edgecolor="none")
plt.title("ToT distribution")
plt.xlabel("ToT [ns]")



# Filename: test_mc.py
# pylint: disable=locally-disabled,C0111,R0904,C0103

from km3pipe.testing import TestCase
from km3pipe.mc import geant2pdg, pdg2name

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
__status__ = "Development"


class TestMc(TestCase):
    def test_geant2pdg(self):
        self.assertEqual(22, geant2pdg(1))
        self.assertEqual(-13, geant2pdg(5))

    def test_geant2pdg_returns_0_for_unknown_particle_id(self):
        self.assertEqual(0, geant2pdg(-999))

    def test_pdg2name(self):
        self.assertEqual("mu-", pdg2name(13))
        self.assertEqual("nu(tau)~", pdg2name(-16))

    def test_pdg2name_returns_NA_for_unknown_particle(self):
        self.assertEqual("N/A", pdg2name(0))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# vim: ts=4 sw=4 et
"""
==================
Reading Timeslices
==================

This examples show how to access L0, L1, L2 and SN timeslice streams using
km3pipe. Note that you need an activated Jpp environment and jppy installed
(http://git.km3net.de/km3py/jppy.git).

Both are available in Lyon, wherebei ``jppy`` is included in the group
python installation.

A general note on timeslice streams: in older Jpp versions there was only
one timeslice stream present in the files. This was later separated into
L0, L1, L2 and SN streams. A convenient tool to check the availability of
streams is ``JPrintTree -f FILENAME``::

    > JPrintTree -f KM3NeT_00000014_00004451.root
    KM3NeT_00000014_00004451.root
    KM3NET_TIMESLICE     KM3NETDAQ::JDAQTimeslice     2808  8416 [MB]
    KM3NET_EVENT         KM3NETDAQ::JDAQEvent              70     0 [MB]
    KM3NET_SUMMARYSLICE  KM3NETDAQ::JDAQSummaryslice     2808     4 [MB]

Here is an example with split timeslice streams::

    > JPrintTree -f KM3NeT_00000029_00003242.root
    KM3NeT_00000029_00003242.root
    KM3NET_TIMESLICE     KM3NETDAQ::JDAQTimeslice          0      0 [MB]
    KM3NET_TIMESLICE_L0  KM3NETDAQ::JDAQTimesliceL0        0      0 [MB]
    KM3NET_TIMESLICE_L1  KM3NETDAQ::JDAQTimesliceL1     5390    319 [MB]
    KM3NET_TIMESLICE_L2  KM3NETDAQ::JDAQTimesliceL2        0      0 [MB]
    KM3NET_TIMESLICE_SN  KM3NETDAQ::JDAQTimesliceSN   107910    162 [MB]
    KM3NET_EVENT         KM3NETDAQ::JDAQEvent          21445     24 [MB]
    KM3NET_SUMMARYSLICE  KM3NETDAQ::JDAQSummaryslice  107910    109 [MB]


To access the Supernova Timeslice hits, you pass ``stream=SN`` to the
``TimeslicePump``, as seen here:

"""

ROOT_FILENAME = "KM3NeT_00000014_00004451.root"

import km3pipe as kp

pump = kp.io.jpp.TimeslicePump(filename=ROOT_FILENAME, stream="SN")
for blob in pump:
    hits = blob["TSHits"]

####################################################################
# The timeslice pump is used to convert the timeslice objects in
# the ROOT file into numpy recarrays. We explicitly set the stream
# to an empty string, since we are opening an older file.
#
# Here is how to access the hits.

pump = kp.io.jpp.TimeslicePump(filename=ROOT_FILENAME, stream="")
for blob in pump:
    hits = blob["TSHits"]



#!/usr/bin/env python
from km3pipe import Module, Pipeline
from km3pipe.io import HDF5Pump


class Printer(Module):
    def process(self, blob):
        print(blob["HitTable"]["dom_id"])
        return blob


FILENAME = "/Users/tamasgal/Data/KM3NeT/DU-2/KM3NeT_00000007_00001597.root.h5"

pipe = Pipeline()
pipe.attach(HDF5Pump, filename=FILENAME)
pipe.attach(Printer)
pipe.drain(5)
